"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import logging
import multiprocessing
import os

from _socket import gethostname

from paho.mqtt import client as mqtt
from paho.mqtt.client import MQTTv5, MQTT_CLEAN_START_FIRST_ONLY

from pyDE1.config import config


def send_single_message(subtopic: str, msg: str,
                        logger: logging.Logger):

    client_logger = logger.getChild('MQTTClient')
    client_logger.setLevel(logging.INFO)

    try:
        pname = multiprocessing.current_process().name
    except AttributeError:
        pname = __name__

    mqtt_client = mqtt.Client(
        client_id="{}@{}[{}]{}_ssm".format(
            config.mqtt.CLIENT_ID_PREFIX,
            gethostname(),
            os.getpid(),
            pname,
        ),
        clean_session=None,  # Required for MQTT5
        userdata=None,
        protocol=MQTTv5,
        transport=config.mqtt.TRANSPORT,
    )

    if config.mqtt.USERNAME is not None:
        logger.info(f"Connecting MQTT with username '{config.mqtt.USERNAME}'")
        mqtt_client.username_pw_set(
            username=config.mqtt.USERNAME,
            password=config.mqtt.PASSWORD
        )

    if config.mqtt.TLS:
        mqtt_client.tls_set(ca_certs=config.mqtt.TLS_CA_CERTS,
                            certfile=config.mqtt.TLS_CERTFILE,
                            keyfile=config.mqtt.TLS_KEYFILE,
                            cert_reqs=config.mqtt.TLS_CERT_REQS,
                            tls_version=config.mqtt.TLS_VERSION,
                            ciphers=config.mqtt.TLS_CIPHERS)

    mqtt_client.connect(host=config.mqtt.BROKER_HOSTNAME,
                        port=config.mqtt.BROKER_PORT,
                        keepalive=config.mqtt.KEEPALIVE,
                        bind_address="",
                        bind_port=0,
                        clean_start=MQTT_CLEAN_START_FIRST_ONLY,
                        properties=None)

    mqtt_client.publish(
        config.mqtt.TOPIC_ROOT + '/' + subtopic,
        msg,
        qos=0,
        retain=True,
    )

    mqtt_client.disconnect()


"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import enum
import logging
import time
import warnings

from typing import (
    Optional, Union, NamedTuple, Callable, Type, TypedDict, Literal
)

import bleak
import bleak.exc
from bleak import BleakClient
from bleak.backends.device import BLEDevice

from pyDE1.exceptions import DE1NoAddressError
from pyDE1.utils import EventReadOnly

try:
    from pyDE1.utils import call_str
except ImportError:
    def call_str(*args, **kwargs) -> str:
        return '(call_str not implemented)'

try:
    from bleak.backends.bluezdbus.scanner import BlueZScannerArgs
except ImportError:
    pass
try:
    from bleak.backends.winrt.client import WinRTClientArgs
except ImportError:
    class WinRTClientArgs(TypedDict, total=False):
        address_type: Literal["public", "random"]
        use_cached_services: bool
#
#   File "[...]]/bleak/backends/winrt/client.py", line 22, in <module>
#     from bleak_winrt.windows.devices.bluetooth import (
# ModuleNotFoundError: No module named 'bleak_winrt'

from pyDE1.lock_logger import LockLogger

IN_PROGRESS_HOLDOFF = 0.2 # seconds, works, 0.1 seemed too short

class CaptureRequest (enum.Enum):
    CAPTURE = 'C'
    RELEASE = 'R'
    CANCEL  = 'X'


class CaptureQueue (NamedTuple):
    connected:  Optional[CaptureRequest]
    pending:    Optional[CaptureRequest]
    target:     Optional[CaptureRequest]

    def __str__(self):
        try:
            retval =  '{}({}, {}, {})'.format(
                self.__class__.__name__,
                self.connected.name if self.connected else None,
                self.pending.name if self.pending else None,
                self.target.name if self.target else None,
            )
        except AttributeError:
            retval = repr(self)

        return retval


def cq_from_code(code: str) -> CaptureQueue:
    tt = {
        'C': CaptureRequest.CAPTURE,
        'R': CaptureRequest.RELEASE,
        'X': CaptureRequest.CANCEL,
        'N': None,
    }
    return CaptureQueue(
        connected=tt[code[0].upper()],
        pending=tt[code[1].upper()],
        target=tt[code[2].upper()],
    )


def cq_to_code(cq: CaptureQueue) -> str:
    retval = ''
    for attr in ('connected', 'pending', 'target'):
        try:
            c = getattr(cq, attr).value
        except AttributeError:
            c = 'N'
        retval += c
    return retval


def task_for_log(t: asyncio.Task) -> str:
    try:
        name = t.get_name()
        state = 'Running'
        if t.cancelled():
            state = 'Cancelled'
        elif t.done():
            state = 'Done'
        retval = f"<Task '{name}', {state}, {t.get_coro().cr_code.co_name}()>"
    except Exception:
        retval = str(t)
    return retval


class ManagedBleakClient (BleakClient):
    """
    Implements BleakClient

    Requires bleak 0.18.1 or later which introduced
    concrete BleakClient and self._backend
    """

    def __init__(self,
                 address_or_ble_device: Optional[Union[BLEDevice, str]],
                 disconnected_callback: Optional[
                     Callable[[BleakClient], None]] = None,
                 *,
                 timeout: float = 10.0,
                 winrt: Optional[WinRTClientArgs] = None,
                 backend: Optional[Type[BleakClient]] = None,
                 on_change_callback: Optional[
                     Callable[['ManagedBleakClient',
                               CaptureQueue,
                               CaptureQueue], None]] = None,
                 logger: Optional[logging.Logger] = None,
                 **kwargs,
                 ):

        if backend is not None:
            raise ValueError(
                f"{self.__class__.__name__} manages its own backends "
                f"through replacement, omit backend= from init call")

        if winrt is None:
            winrt = {}

        # Assigning logger needs to be ahead of generate_disconnected_callback
        if logger is None:
            logger = logging.getLogger(self.__class__.__name__)
            logger.warning(f"Using fall-back logger {logger}")
        self.logger = logger

        self._legacy_disconnected_callback = disconnected_callback
        self._on_change_callback = on_change_callback

        # Retain init params for potentially replacing backend
        self._init_winrt = winrt
        self._init_kwargs = kwargs

        self._change_address_lock = asyncio.Lock()

        if address_or_ble_device is None:
            address_or_ble_device = ''

        # NB: Replicate any changes here in change_address()
        super(ManagedBleakClient, self).__init__(
                address_or_ble_device=address_or_ble_device,
                disconnected_callback=self._generate_disconnected_callback(),
                timeout=timeout,
                winrt=winrt,
                backend=backend,
                **kwargs,
        )

        self._capture_queue_lock = asyncio.Lock()
        self.__capture_queue: CaptureQueue = CaptureQueue(None, None, None)

        self._event_captured = asyncio.Event()
        self._event_released = asyncio.Event()
        self._event_no_pending = asyncio.Event()
        self._event_connected = asyncio.Event()
        self._event_disconnected = asyncio.Event()

        # Cache wrapped versions
        self._event_captured_ro = EventReadOnly(self._event_captured)
        self._event_released_ro = EventReadOnly(self._event_released)
        self._event_no_pending_ro = EventReadOnly(self._event_no_pending)
        self._event_connected_ro = EventReadOnly(self._event_connected)
        self._event_disconnected_ro = EventReadOnly(self._event_disconnected)


        self._retry_wait_event = asyncio.Event()
        self._retry_wait_task: Optional[asyncio.Task] = None
        # Using an event allow retries to resume immediately
        # without fancy scheduling and rescheduling
        self._retry_since = None
        #
        self._retry_start_initial_delay = 300 # seconds
        self._retry_initial_delay = 15 # seconds
        self._retry_start_long_delay = 1800  # seconds
        self._retry_long_delay = 60 # seconds

        self._pending_task: Optional[asyncio.Task] = None

        self._capture_release_done_callback = \
            self._generate_capture_release_done_callback()

        self._reset_all_unsafe()

    ###
    ### Public API
    ###

    async def change_address(
            self, new_addr: Optional[Union[BLEDevice, str]] = None) -> bool:
        """
        If address is the same, returns immediately
        If the address is different, will
          * Disconnect existing client
          * Create a new backend with the parameters saved from __init__()
            but with the new address and save the new address

        If None or the empty string is passed, disconnects and "forgets" address

        NB: Other actions aren't explicitly locked out during an address change

        :returns    bool    True if the address has been changed
        """
        if new_addr is None:
            new_addr = ''

        if isinstance(new_addr, BLEDevice):
            check_addr = new_addr.address
        else:
            check_addr = new_addr
        if check_addr == self.address:
            self.logger.info(
                f"change_address({check_addr}) would not change address. "
                "No action taken.")
            return False

        ll = LockLogger(self._change_address_lock, 'ChangeAddress').check()
        async with self._change_address_lock:
            ll.acquired()
            # NB: Other actions aren't explicitly locked out
            # NB: Go directly to _queue_request as release or request_release
            #     would try to acquire _change_address_lock and hang
            await self._queue_request(CaptureRequest.RELEASE)
            await self.event_released.wait()

            # NB: This makes some assumptions about bleak internals
            current_disconnected_callback = self._backend._disconnected_callback
            current_timeout = self._backend._timeout

            new_bleak_client = BleakClient(
                address_or_ble_device=check_addr,
                disconnected_callback=current_disconnected_callback,
                timeout=current_timeout,
                winrt=self._init_winrt,
                backend=None,
                **self._init_kwargs,
            )
            self._backend = new_bleak_client._backend
            self._reset_all_unsafe()

        ll.released()
        return True

    async def connect(self, **kwargs) -> bool:
        """
        Intercept connect and replace with "capture()"

        super:
            Connect to the specified GATT server.

            Args:
                **kwargs: For backwards compatibility - should not be used.

            Returns:
                Always returns ``True`` for backwards compatibility.
        """
        self.logger.warning(
            f"Prefer {self.__class__.__name__}.capture() to .connect()"
        )
        await self.capture()
        return True

    async def disconnect(self) -> bool:
        """
        Intercept disconnect and replace with "release()"

        super:

            Disconnect from the specified GATT server.

            Returns:
                Always returns ``True`` for backwards compatibility.
        """
        self.logger.warning(
            f"Prefer {self.__class__.__name__}.release() to .disconnect()"
        )
        await self.release()
        return True

    async def capture(self, timeout: Optional[float] = None) -> bool:
        """
        Request capture and wait up to timeout
        Default is self._backend._timeout

        :return: is_captured
        """
        await self.request_capture()
        if timeout is None:
            timeout = self._backend._timeout
        await asyncio.wait_for(self.event_captured.wait(), timeout=timeout)
        return self.is_captured

    async def release(self, timeout: Optional[float] = None) -> bool:
        """
        Request capture and wait up to timeout
        Default is self._backend._timeout

        :return: is_released
        """
        await self.request_release()
        await asyncio.wait_for(self.event_released.wait(), timeout=timeout)
        return self.is_released

    async def request_capture(self):
        """
        Request capture and "immediately" return
        """
        if not self.address:
            raise DE1NoAddressError(f"{self}.address: '{self.address}'")
        ll = LockLogger(self._change_address_lock, "ChangeAddress").check()
        async with self._change_address_lock:
            ll.acquired()
            await self._queue_request(CaptureRequest.CAPTURE)
        ll.released()

    async def request_release(self):
        """
        Request release and "immediately" return
        """
        ll = LockLogger(self._change_address_lock, "ChangeAddress").check()
        async with self._change_address_lock:
            ll.acquired()
            await self._queue_request(CaptureRequest.RELEASE)
        ll.released()

    @property
    def event_captured(self) -> EventReadOnly:
        """
        Returns an asyncio.Event that is set when the device is captured
        """
        return self._event_captured_ro

    @property
    def event_released(self) -> EventReadOnly:
        """
        Returns an asyncio.Event that is set when the device is released
        """
        return self._event_released_ro

    @property
    def event_no_pending(self) -> EventReadOnly:
        """
        Returns an asyncio.Event that is set when there are no pending tasks
        on the capture/release queue
        """
        return self._event_no_pending_ro

    @property
    def event_connected(self) -> EventReadOnly:
        """
        Returns an asyncio.Event that is set when there are no pending tasks
        on the capture/release queue
        """
        return self._event_connected_ro

    @property
    def event_disconnected(self) -> EventReadOnly:
        """
        Returns an asyncio.Event that is set when there are no pending tasks
        on the capture/release queue
        """
        return self._event_disconnected_ro

    # Supplied by superclass
    # @property
    # def is_connected(self) -> self._DeprecatedIsConnectedReturn:
    # (not bool)

    @property
    def is_captured(self) -> bool:
        return self._event_captured.is_set()

    @property
    def is_released(self) -> bool:
        return self._event_released.is_set()

    @property
    def connectivity_task_pending(self) -> bool:
        """
        At least at this time, is_captured and is_released
        are "at the moment", even if there is a task running
        that may change the connectivity.
        """
        return self._capture_queue.pending is not None

    @property
    def active_request(self) -> Optional[CaptureRequest]:
        """
        Returns the current target/goal for capture/release
        """
        return self._capture_queue.target

    @property
    def on_change_callback(self):
        return self._on_change_callback
    
    @on_change_callback.setter
    def on_change_callback(self, val: Callable):
        """
        Callback is called any time the capture queue changes with
        (mbc: ManagedBleakClient, old: CaptureQueue, new: CaptureQueue)
        """
        self._on_change_callback = val

    def set_disconnected_callback(
        self, callback: Optional[Callable[[BleakClient], None]], **kwargs
    ) -> None:
        """Set the disconnect callback.

        .. deprecated:: 0.17.0
            This method will be removed in a future version of Bleak.
            Pass the callback to the :class:`BleakClient` constructor instead.

        Args:
            callback: callback to be called on disconnection.

        """
        warnings.warn(
            "This method will be removed future version, pass the callback "
            f"to the {self.__class__.__name__} constructor instead.",
            FutureWarning,
            stacklevel=2,
        )
        self._legacy_disconnected_callback = callback


    ###
    ### End of public API
    ###

    def _reset_all_unsafe(self):
        # self.logger.info(f'_reset_all_unsafe()')
        self.logger.debug(f'_reset_all_unsafe() {call_str()}')
        self._event_captured.clear()
        self._event_released.clear()
        self._event_no_pending.set()
        self._event_disconnected.clear()
        self._event_connected.clear()
        self._retry_reset()
        if self._pending_task is not None:
            self._pending_task.cancel()
            # NB: It may still come back to haunt
        # Note accessing double-underscore version
        # and then calling the on-change method
        self.__capture_queue = CaptureQueue(None, None, None)
        if self._on_change_callback is not None:
            self._on_change_callback(self,
                                     self._capture_queue,
                                     self._capture_queue)

    @property
    def _capture_queue(self):
        return self.__capture_queue

    @_capture_queue.setter
    def _capture_queue(self, new_cq: CaptureQueue):
        old_cq = self.__capture_queue
        self.__capture_queue = new_cq
        if new_cq != old_cq and self._on_change_callback is not None:
            logger = self.logger.getChild('CQ')
            logger.debug(
                'Calling {} with {} => {}'.format(
                    self._on_change_callback.__name__,
                    cq_to_code(old_cq),
                    cq_to_code(new_cq)
                ))
            self._on_change_callback(self, old_cq, new_cq)


    async def _queue_request(self, request: CaptureRequest) -> bool:
        if request not in (CaptureRequest.CAPTURE,
                           CaptureRequest.RELEASE):
            raise ValueError(
                "Request must be CaptureState.CAPTURE or .RELEASE, "
                f"not {request}")
        ll = LockLogger(self._capture_queue_lock, 'CaptureQueue').check()
        async with self._capture_queue_lock:
            ll.acquired()
            retval = self._maybe_initiate_action_have_lock(request=request)
        ll.released()
        return retval

    def _maybe_initiate_action_have_lock(self,
                            request: Optional[CaptureRequest] = None) -> bool:
        """
        Given the current _capture_queue, evaluate the target, current,
        and pending. If needed, initiate a new task or cancel the pending.
        Update _capture_queue

        If request is None, use the existing target

        Returns True if an action was taken

        NB: Only call while holding _capture_queue_lock
        """
        assert self._capture_queue_lock.locked(), \
                f"{self}._capture_queue_lock was not locked "

        logger = self.logger.getChild('Initiate')

        next_cq_action = None
        new_action_taken = False

        cq = self._capture_queue
        if request is None:
            request = cq.target

        if request is None:
            logger.warning(
                f"No request, no change from {self._capture_queue}")
            return False

        if cq.connected == request:

            if cq.pending is None:
                next_cq_action = None

            else:
                next_cq_action = CaptureRequest.CANCEL

        else:  # cq.current != request

            if cq.pending is None:
                next_cq_action = request

            elif cq.pending == request:
                next_cq_action = request

            else:
                next_cq_action = CaptureRequest.CANCEL

        if cq.pending != next_cq_action:
            self._start_request_with_lock(next_cq_action)
            new_action_taken = True

        self._capture_queue = CaptureQueue(
            connected=cq.connected,
            pending=next_cq_action,
            target=request,
        )

        if next_cq_action== CaptureRequest.CAPTURE:
            if not self._retry_is_active:
                self._retry_start()
            else:
                pass
            self._retry_set_timer()
        else:
            self._retry_reset()

        return new_action_taken

    def _start_request_with_lock(self, req: CaptureRequest):
        if req == CaptureRequest.CAPTURE:
            self._start_capture_with_lock()
        elif req == CaptureRequest.RELEASE:
            self._start_release_with_lock()
        elif req == CaptureRequest.CANCEL:
            self._start_cancel_with_lock()
        else:
            raise ValueError(f"Unrecognized request: {req}")

    def _start_capture_with_lock(self):
        # TODO: build this out
        self.logger.debug("Start capture")
        self._retry_set_timer()
        t = asyncio.create_task(self._backend_connect_after_retry_wait_event())
        self._pending_task = t
        t.add_done_callback(self._capture_release_done_callback)
        self.logger.debug(f"task: {task_for_log(t)}")
        return t

    def _start_release_with_lock(self):
        self.logger.info("Start release")
        t = asyncio.create_task(self._backend.disconnect())
        self._pending_task = t
        t.add_done_callback(self._capture_release_done_callback)
        self.logger.debug(f"task: {task_for_log(t)}")
        return t

    def _start_cancel_with_lock(self):
        self.logger.info("Start cancel")
        if (pt := self._pending_task) is not None:
            self.logger.debug(f"Cancel request: {task_for_log(pt)}")
            pt.cancel()
        else:
            self.logger.warning('No pending task to cancel')

    # Retry approach:
    # Only for connect/capture
    # Use _retry_wait_event so that it can be immediately released
    # if it is requested again

    async def _backend_connect_after_retry_wait_event(self):
        # TODO: This should be cleaner on timeout
        await self._retry_wait_event.wait()
        # NB: Not clearing here has the advantage for testing
        #     that the event is always set after _backend.connect()
        # self._retry_wait_event.clear()
        try:
            await self._backend.connect()
        except asyncio.CancelledError:
            self.logger.info("connect retry CancelledError caught, pass")
            pass
        except asyncio.TimeoutError:
            self.logger.info("connect retry TimeoutError caught, pass")
            pass
        except bleak.exc.BleakDeviceNotFoundError as e:
            # bleak 0.19.0 and later
            e: bleak.exc.BleakDeviceNotFoundError
            self.logger.info(f"Seemingly stale device, resetting: {e}")
            self._backend._device_path = None
            await self._backend_connect_after_retry_wait_event()
        except bleak.exc.BleakDBusError as e:
            # Here we go again, parsing messages to determine *which* exception
            if e.args[0] == 'org.bluez.Error.InProgress':
                self.logger.info(
                    f"connect retry caught {e}, delaying this one a bit")
                await asyncio.sleep(IN_PROGRESS_HOLDOFF)
                await self._backend_connect_after_retry_wait_event()
            else:
                self.logger.exception(
                    'Failed to connect(), unrecognized exception.')
                raise
        except Exception:
            self.logger.exception(
                'Failed to connect(), unrecognized exception.')
            raise


    @property
    def _retry_is_active(self):
        return self._retry_since is not None

    def _retry_reset(self):
        self.logger.debug("Resetting retry timer")
        if self._retry_wait_task is not None:
            self._retry_wait_task.cancel()
            self._retry_wait_task = None
        self._retry_since = None
        self._retry_wait_event.clear()

    def _retry_start(self):
        self._retry_reset()
        self.logger.debug("Starting retry timer")
        self._retry_since = time.time()

    def _retry_delay(self) -> float:
        if self._retry_since is None:
            retval = 0
        else:
            dt = time.time() - self._retry_since
            if dt < self._retry_start_initial_delay:
                retval = 0
            elif self._retry_start_initial_delay <= dt < self._retry_start_long_delay:
                retval = self._retry_initial_delay
            else:
                retval = self._retry_long_delay
        return retval

    def _retry_set_timer(self):
        how_long = self._retry_delay()
        if how_long == 0:
            self._retry_wait_event.set()
        else:
            self._retry_wait_event.clear()
            self._retry_wait_task = asyncio.create_task(
                self._retry_waiter(how_long),
                name=f"_retry_waiter({how_long})")
            self.logger.info(
                f"Connect holdoff of {how_long} seconds "
                + task_for_log(self._retry_wait_task))

    # Potentially use loop.call_later
    # returns an asyncio.TimerHandle
    # th.cancel()
    # th.when() float in seconds, absolute

    async def _retry_waiter(self, how_long: float):
        assert how_long > 0
        try:
            t0 = time.time()
            await asyncio.sleep(how_long)
            dt = time.time() - t0
            self._retry_wait_event.set()
        except asyncio.CancelledError:
            pass
        finally:
            self._retry_wait_task = None

    # https://docs.python.org/3/library/asyncio-task.html#asyncio.Task.add_done_callback
    # https://docs.python.org/3/library/asyncio-future.html#asyncio.Future.add_done_callback
    # add_done_callback(callback, *, context=None)
    #
    # The callback is called with the Future object as its only argument.
    #
    # An optional keyword-only context argument allows specifying a custom
    # contextvars.Context for the callback to run in.
    # The current context is used when no context is provided.

    def _generate_capture_release_done_callback(self):
        # Try without capturing self explicitly
        logger = self.logger.getChild('DoneCB')

        def capture_release_done_callback(fut: asyncio.Future):
            logger.debug(f"Entering done callback {task_for_log(fut)}")

            try:
                fut.result()
            except asyncio.CancelledError:
                pass
            except Exception:
                logger.debug(
                    f"client: {self._backend} addr type: {type(self._backend.address)}")
                logger.exception(f"Exception from {fut}")
                raise

            if (wt := self._retry_wait_task) is not None:
                wt.cancel()
                logger.info(
                    f"In {task_for_log(fut)}, canceled {task_for_log(wt)}")
            asyncio.create_task(
                self._capture_release_done_callback_async(
                    done_callback_from=fut))
            # logger.info(f"Leaving callback {task_for_log(fut)}")

        return capture_release_done_callback

    async def _capture_release_done_callback_async(self,
            done_callback_from: asyncio.Future):
        logger = self.logger.getChild('DoneCB.A')
        logger.debug(
            f"Entering async done callback {task_for_log(done_callback_from)}")
        if done_callback_from.cancelled():
            logger.debug(
                f"Done callback reports cancelled {task_for_log(done_callback_from)}")

        ll = LockLogger(lock=self._capture_queue_lock,
                        name='CaptureQueue').check()
        async with self._capture_queue_lock:
            ll.acquired()

            # Sanity checks
            if self._capture_queue.pending is None:
                logger.warning(
                    f"Nothing pending, but got {done_callback_from}.")
            if done_callback_from != self._pending_task:
                logger.error(
                    f"Pending {self._pending_task}, "
                    f"but got {done_callback_from}. "
                    "Removing anyway.")

            # Update to the new result
            self._pending_task = None
            cval = CaptureRequest.CAPTURE if self._backend.is_connected \
                else CaptureRequest.RELEASE
            self._capture_queue = CaptureQueue(
                    connected=cval,
                    pending=None,
                    target=self._capture_queue.target,
            )
            # The update with any needed action
            self._update_queue_and_events_have_lock()

        ll.released()
        # logger.info(
        #     f"Leaving async callback {task_for_log(done_callback_from)}")

    def _update_queue_and_events_have_lock(self):
        """
        Called in done_callback from a task to change connectivity
        as well as from disconnected_callback.

        NB: Only call while holding _capture_queue_lock
        """
        assert self._capture_queue_lock.locked(), \
                f"{self}._capture_queue_lock was not locked "

        logger = self.logger.getChild('Update')

        cq_on_entry = self._capture_queue
        if self.is_connected:
            current = CaptureRequest.CAPTURE
        else:
            current = CaptureRequest.RELEASE

        self._capture_queue = CaptureQueue(connected=current,
                                           pending=cq_on_entry.pending,
                                           target=cq_on_entry.target)
        logger.debug(
            "As {}connected, updated from {} to {} ".format(
                '' if self.is_connected else 'dis',
                cq_to_code(cq_on_entry),
                cq_to_code(self._capture_queue)
            ))

        self._maybe_initiate_action_have_lock(request=cq_on_entry.target)

        logger.debug(
            'After _maybe_initiate: was {} now {}'.format(
                cq_to_code(cq_on_entry),
                cq_to_code(self._capture_queue)
            ))

        #
        # Update the events
        #

        cq = self._capture_queue

        if cq.connected == CaptureRequest.CAPTURE:
            self._event_disconnected.clear()
            self._event_released.clear()
            self._event_connected.set()
            if cq.pending is None:
                self._event_captured.set()
                self._event_no_pending.set()
            else:
                self._event_captured.clear()
                self._event_no_pending.clear()
            logger.debug("Events set/cleared for CaptureRequest.CAPTURE")

        elif cq.connected == CaptureRequest.RELEASE:
            self._event_connected.clear()
            self._event_captured.clear()
            self._event_disconnected.set()
            if cq.pending is None:
                self._event_released.set()
                self._event_no_pending.set()
            else:
                self._event_released.clear()
                self._event_no_pending.clear()
            logger.debug("Events set/cleared for CaptureRequest.RELEASE")

        elif cq.connected is None:
            pass

        else:
            raise ValueError(
                "Update events: Current should only be CAPTURE or RELEASE, "
                f"not {cq}"
            )

    def _generate_disconnected_callback(self):
        # Try without capturing self explicitly
        logger = self.logger.getChild('DiscCB')

        # TODO: Python 3.11 introduces typing.Self
        def disconnected_callback(client: 'ManagedBleakClient'):
            logger.debug(f"Disconnected callback, create async task {client}")
            asyncio.create_task(
                self._disconnected_callback_async(disconnected_from=client))
            # logger.info(f"Leaving callback {client}")

        return disconnected_callback

    async def _disconnected_callback_async(self,
                                disconnected_from: 'ManagedBleakClient'):
        logger = self.logger.getChild('DiscCB.A')
        logger.debug(
            f"Entering async disconnected callback {disconnected_from}")

        ll = LockLogger(lock=self._capture_queue_lock,
                        name='CaptureQueue').check()
        async with self._capture_queue_lock:
            ll.acquired()
            # Update to the new result
            cval = CaptureRequest.CAPTURE if self._backend.is_connected \
                else CaptureRequest.RELEASE
            self._capture_queue = CaptureQueue(
                connected=cval,
                pending=self._capture_queue.pending,
                target=self._capture_queue.target,
            )
            # Then update with any needed action
            self._update_queue_and_events_have_lock()

        ll.released()
        if self._legacy_disconnected_callback is not None:
            logger.info("Calling legacy disconnected_callback")
            self._legacy_disconnected_callback(disconnected_from)
            logger.info("Returned from calling legacy disconnected_callback")
        # logger.info(f"Leaving async callback {disconnected_from}")




"""
Copyright © 2022 Mimoja. All Rights Reserved.
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

Content by Mimoja used under grant of GPL-v3.0-only at
https://github.com/Mimoja/pyDE1/blob/stable/src/pyDE1/scale/felicita_arc.py

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import enum
import time

import pyDE1
from pyDE1.scale.events import ScaleWeightUpdate
from pyDE1.scale.generic_scale import register_scale_class, GenericScale


logger = pyDE1.getLogger('Scale.FelicitaArc')


@register_scale_class
class FelicitaArc (GenericScale):

    _supports_prefixes = ['FelicitaArc']

    def __init__(self):
        super(FelicitaArc, self).__init__()
        self._adopt_sync()

    def _adopt_sync(self):
        self._nominal_period = 0.1  # seconds per sample
        self._minimum_tare_request_interval = 2.5 * self._nominal_period
        self._sensor_lag = 0.45  # seconds, including all transit delays
        self._tare_timeout = 1.0  # seconds until considered coincidence
        self._tare_threshold = 0.05  # grams, within this, considered "at zero"

    async def _adopt_class(self):
        self._adopt_sync()

    async def _leave_class(self):
        pass   # Nothing other than common attributes set

    async def start_sending_weight_updates(self):
        await self._bleak_client.start_notify(
            Characteristic.MAIN.cuuid,
            self._weight_update_hander)
        logger.info("Sending weight updates")

    async def stop_sending_weight_updates(self):
        await self._bleak_client.stop_notify(
            Characteristic.MAIN.cuuid)
        logger.info("Stopped weight updates")

    def is_sending_weight_updates(self):
        return NotImplementedError

    async def _tare_internal(self):
        await self.send_command(Command.TARE)
        logger.info("Internal tare sent")

    async def current_weight(self):
        return None

    async def display_on(self):
        pass
        logger.info("Display on")

    async def display_off(self):
        pass
        logger.info("Display off")

    async def set_grams(self):
        pass
        logger.info("Grams selected")

    @property
    def supports_button_press(self):
        return False

    async def send_command(self, command: "Command"):
        await self._bleak_client.write_gatt_char(
            command.cuuid,
            command.value,
            response=self._write_gatt_char_response)

    async def _weight_update_handler(self, sender, data):

        try:
            now = time.time()
            if len(data) < 9:
                return

            # data[0:2] is a header
            sign = data[2]
            weight = int(data[3:]) / 100.0

            if chr(sign) == '-':
                weight *= -1

            self._update_scale_time_estimator(now)

            await self.event_weight_update.publish(
                ScaleWeightUpdate(
                    arrival_time=now,
                    scale_time=self._scale_time_from_latest_arrival(now),
                    weight=weight
                ))
        except Exception as e:
            logger.exception(e)
            raise e


class Characteristic(enum.Enum):

    MAIN =    'FFE1'  # RWN

    @property
    def cuuid(self):
        return f"0000ffe1-0000-1000-8000-00805f9b34fb"


class Command(enum.Enum):

    TARE = b'\x54'

    TIMER_ZERO = b'\x43'
    TIMER_START = b'\x52'
    TIMER_STOP = b'\x53'

    @property
    def cuuid(self):
        return Characteristic.MAIN.cuuid



"""
Utility to catch and log task exceptions

Source: https://quantlane.com/blog/ensure-asyncio-task-exceptions-get-logged/

    HOW WE MADE THIS ERROR HANDLING REUSABLE

    Since long-running tasks are a common pattern in our code we created
    a utility function to set up all these error handlers for us.
    This is our task_logger.py module.
    It is MIT-licenced, so feel free to use and modify for your own needs:

    The full example code taking advantage of this follows.
    We simply use task_logger.create_task on line 22
    and provide a bit of context to it.

 1 import asyncio
 2 import logging
 3
 4 import task_logger
 5
 6
 7 async def problem() -> None:
 8     await asyncio.sleep(1)
 9     logging.warning('Going to raise an exception now!')
10     raise RuntimeError('Something went wrong')
11
12
13 if __name__ == '__main__':
14     logging.basicConfig(
15         format = '▸ %(asctime)s.%(msecs)03d %(filename)s:%(lineno)d %(levelname)s %(message)s',
16         level = logging.INFO,
17         datefmt = '%H:%M:%S',
18     )
19     loop = asyncio.get_event_loop()
20     logging.info('Creating the problem task')
21     logger = logging.getLogger('task_logger')
22     task = task_logger.create_task(problem(), logger = logger, message = 'Task raised an exception', loop = loop)
23     logging.info('Created task = %r', task)
24     logging.info('Running the loop')
25     try:
26         loop.run_forever()
27     except KeyboardInterrupt:
28         logging.info('Closing the loop')
29         loop.close()
30     logging.info('Shutting down')

        Written by Vita Smid on June 26, 2020.
"""
from typing import Any, Awaitable, Optional, TypeVar, Tuple

import asyncio
import functools
import logging


T = TypeVar('T')


def create_task(
    coroutine: Awaitable[T],
    *,
    logger: logging.Logger,
    message: str,
    message_args: Tuple[Any, ...] = (),
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> 'asyncio.Task[T]':  # This type annotation has to be quoted for Python < 3.9, see https://www.python.org/dev/peps/pep-0585/
    '''
    This helper function wraps a ``loop.create_task(coroutine())`` call and ensures there is
    an exception handler added to the resulting task. If the task raises an exception it is logged
    using the provided ``logger``, with additional context provided by ``message`` and optionally
    ``message_args``.
    '''
    if loop is None:
        loop = asyncio.get_running_loop()
    task = loop.create_task(coroutine)
    task.add_done_callback(
        functools.partial(_handle_task_result, logger = logger, message = message, message_args = message_args)
    )
    return task


def _handle_task_result(
    task: asyncio.Task,
    *,
    logger: logging.Logger,
    message: str,
    message_args: Tuple[Any, ...] = (),
) -> None:
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    # Ad the pylint ignore: we want to handle all exceptions here so that the result of the task
    # is properly logged. There is no point re-raising the exception in this callback.
    except Exception:  # pylint: disable=broad-except
        logger.exception(message, *message_args)




"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

New information on protocol and values based packet capture of
a Lunar 2021 and "acaia Coffee" iOS in Jan/Feb 2022 by Jeff Kletsky

---

Naming of some elements of MessageType and EventType after
https://github.com/ntoto/ACAIAScale_Arduino/blob/master/Scale.cpp
As such, these names might be considered as

Copyright (c) 2017 Nicolas Pouvesle

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# 
# LUNAR-E01E31 (Lunar 2021 on Firmware 1.00.001 or 1.00.005)
# 
# nRF Connect:
# ------------
# Connected at 13:16:19
# Last resp at 13:16:20
# Timed out at 13:16:26
# 
# Services:
# 49535343-C9D0-CC83-A44A-6FE238D06D33, 
#     49535343-FE7D-4AE5-8FA9-9FAFD205E455, 
#     and Device Information Services (180a)
# 
# Service 49535343-C9D0-CC83-A44A-6FE238D06D33
#     Char 49535343-ACA3-481C-91EC-D85E28A60318 write, notify
#     
# Service 49535343-FE7D-4AE5-8FA9-9FAFD205E455
#     Char 49535343-1E4D-4BD9-BA61-23C647249616 write, notify (WEIGHT)
#     Char 49535343-4C8A-39B3-2F49-511CFF073B7E write, write w/o resp, notify
#     Char 49535343-8841-43F4-A8D4-ECBE34729BB3 write, write w/o resp (COMMAND)
#     
# 2a23    System ID
# 2a24    Model Number String
# 2a25    Serial Number String
# 2a26    Firmware Revision String
# 2a27    Hardware Revision String
# 2a28    Software Revision String
# 2a29    Manufacturer Name String
# 2a2a    IEEE 11073-20601 Regulatory Certification Data List
#

# 2022-12-17 -- available firmware
#   Lunar 2.6.016           (AL006 in hand)
#   Lunar (2021) 1.0.009
#
# Seen on ACAIAL1
#
# 2a80  Age             Write w/o, notify
# 2a19  Battery level   Read, notify
# 2a50  PnP ID          Read -- 0x30-30-31-43-39-37
#
# 2a25  Serial number
# 2a26  Firmware rev
# 2a27  Hardware rev
# 2a28  Software rev
# 2a29  Manufacturer
#

import asyncio
import enum
import logging
import time
from typing import Optional, Union, Coroutine, Awaitable, NamedTuple

import bleak

import pyDE1
import pyDE1.event_manager
from pyDE1.scale.generic_scale import GenericScale, register_scale_class
from pyDE1.scale.events import ScaleWeightUpdate
from pyDE1.supervise import SupervisedTask

from pyDE1.config import config

logger = pyDE1.getLogger('Scale.Acaia')
logger_notify = logger.getChild('Notify')


class InteractionStyle (enum.Enum):
    CLASSIC = 'classic'
    LUNAR = 'lunar'


class CUUID (enum.Enum):
    # "CLASSIC"
    # '2a80' is an Assigned Number in "16-bit UUID Numbers Document", 2022-01-26
    # https://btprodspecificationrefs.blob.core.windows.net/assigned-values/16-bit%20UUID%20Numbers%20Document.pdf
    AGE     = '00002a80-0000-1000-8000-00805f9b34fb'
    # "LUNAR"
    COMMAND = '49535343-8841-43f4-a8d4-ecbe34729bb3'  # write, w/o resp
    WEIGHT  = '49535343-1e4d-4bd9-ba61-23c647249616'  # write, notify
    UNKNOWN = '49535343-4c8a-39b3-2f49-511cff073b7e'  # write, w/o resp, notify

    # These seem to be from https://www.microchip.com/en-us/product/BM70
    # rather than related to the Lunar (running v1.00.001 FW at the time)
    SYSTEM_ID       = '2a23'    # b'\x00\x00\x00\x00\x00\x00\x00\x00'
    MODEL_NUMBER    = '2a24'    # b'BM70'
    SERIAL_NUMBER   = '2a25'    # b'0000'
    FIRMWARE_REV    = '2a26'    # b'009010'
    HARDWARE_REV    = '2a27'    # b'5505 102_LESDK'
    SOFTWARE_REV    = '2a28'    # b'0000'
    MANUFACTURER    = '2a29'    # b'MCHP'
    # This is apparently related to medical devices, like glucose meters
    IEEE_11073_20601 = '2a2a'   # b'\x00\x00\x00\x00\x01\x00\x00\x00'

# When this message is delivered, it seems that the Lunar 2021
# isn't "properly" connected above the Bluetooth layer
# It seems that it will disconnect Bluetooth relatively quickly
# At this time (Feb, 2022), detection of this packet is used
# to preemptively disconnect and reconnect to the Acaia scale.

NOT_REALLY_CONNECTED = bytes.fromhex('efdd 0707 0219 0100 0501 2108')

# May also not be really connected if CUUID.COMMAND is not present


# Header byte pattern, may be multiple messages in single packet
# message[0] and message[1]
# This is intentionally bytes so .startswith() and .find() work
# TODO: Is bytes still needed or would bytearray work?

HEADER = b'\xef\xdd'


# message[2]
class MessageType (enum.IntEnum):
    SYSTEM      = 0x00
    TARE        = 0x04
    INFO        = 0x07
    STATUS      = 0x08
    IDENTIFY    = 0x0b
    EVENT       = 0x0c
    TIMER       = 0x0d

    REQUEST_02  = 0x02
    REQUEST_06  = 0x06
    CONFIG      = 0x0a


# message[3] is length in bytes,
# exclusive of three leading and two trailing bytes
# length includes itself
#
# message[3] can also be a sequence number
# for things like config, status requests, heartbeat, timer, (more?)


# TIMER request is sent with a sequence number
class TimerRequestType (enum.Enum):
    START   = b'\x00'
    RESET   = b'\x01'  # Appears to both stop and zero
    STOP    = b'\x02'


# CONFIG requests are sent with a sequence number
class ConfigType (enum.IntEnum):
    UNITS       = 0x00  # 0x05 oz, 0x02 grams; STATUS[1]
    AUTO_OFF    = 0x01  # 0x00 off, 0x01 5 min, 0x02 10 min,
                        # 0x03 20 min, 0x04 30 min, 0x05 60 min; STATUS[3]
    TOUCHPAD    = 0x02  # 0x78 120 seconds disabled; STATUS[4]
    UNKNOWN     = 0x03
    CAPACITY    = 0x04  # 0x00 1 kg, 0x01 2 kg; STATUS[7]
    BEEP        = 0x05  # 0x00 off, 0x01 on; STATUS[5]


class ConfigUnits (enum.IntEnum):
    OZ = 0x05
    G =  0x02


class ConfigAutoOff (enum.IntEnum):
    NONE   = 0x00
    MIN_5  = 0x01
    MIN_10 = 0x02
    MIN_20 = 0x03
    MIN_30 = 0x04
    MIN_60 = 0x05


class ConfigRange (enum.IntEnum):
    KG_1 = 0x00
    KG_2 = 0x01


class ConfigBeep (enum.IntEnum):
    OFF = 0x00
    ON  = 0x01


class Info (NamedTuple):
    battery: int
    units: ConfigUnits
    unk2: int
    auto_off: ConfigAutoOff
    unk4: int
    beep: ConfigBeep
    range: ConfigRange


# message[4]
class EventType (enum.IntEnum):
    WEIGHT      = 0x05
    # BATTERY     = 0x06
    TIMER       = 0x07
    KEY         = 0x08
    ACK         = 0x0b

    REPLY_06    = 0x06


# message[-2] is even-index checksum
# message[-1] is odd-index checksum
# Each checksum is the least-significant byte of the sum
# over the bytes of the payload (including its length byte)

def checksum(len_payload: Union[bytes, bytearray],
             is_request = False) -> bytearray:
    if not is_request and (len(len_payload)) != len_payload[0]:
        raise ValueError(
            "Inconsistent length byte in len_payload "
            f"{len_payload.hex(sep=' ', bytes_per_sep=-2)}")
    evens = 0
    odds = 0
    for idx in range(0, len(len_payload)):
        if (idx % 2):
            odds += len_payload[idx]
        else:
            evens += len_payload[idx]
    return bytearray((evens & 0xff, odds & 0xff))


def pack_message(message_type: MessageType,
                 payload: Union[bytes, bytearray]) -> bytearray:
    length_byte = 1 + len(payload)
    if length_byte > 255:
        raise ValueError("Unexpectedly long payload "
                         f"{payload.hex(sep=' ', bytes_per_sep=-2)}")
    packed = bytearray(HEADER)
    packed.extend((message_type.value, length_byte))
    packed.extend(payload)
    packed.extend(checksum(packed[3:]))
    return packed


def pack_request(message_type: MessageType,
                 sequence_number: int,
                 payload: Union[bytes, bytearray]) -> bytearray:
    if len(payload) > 2:
        raise ValueError("Unexpectedly long payload "
                         f"{payload.hex(sep=' ', bytes_per_sep=-2)}")
    packed = bytearray(HEADER)
    packed.extend((message_type.value, sequence_number))
    packed.extend(payload)
    packed.extend(checksum(packed[3:], is_request=True))
    return packed


def pack_config(setting_type: ConfigType,
                sequence_number: int,
                setting_enum_instance: Union[
                     ConfigUnits,
                     ConfigAutoOff,
                     ConfigRange,
                     ConfigBeep
                 ]) -> bytearray:
    # Why is this a tuple that is raising "unexpected type"?
    # Probably because it is two int values that aren't seen that way
    payload = bytes((setting_type.value, setting_enum_instance.value))
    return pack_request(MessageType.CONFIG, sequence_number, payload)


class FixedMessage (enum.Enum):
    """
    These messages are not time/situation varying
    and may not follow the `pack_message` format

    Gathered from iOS "acaia Coffee" app on 2022-01-31 and again 2022-12-18
    """
    IDENT   = bytes.fromhex('EFDD 0B2D 2D2D 2D2D 2D2D 2D2D 2D2D 2D2D 2D2D 683B')

    # Follows with
    #   Channel ID: 0x0004  Length: 0x0011 (17) [ 52 0D 00
    #   EF DD 0C 09 00 01 01 02 02 05 03 04 15 ... ]

    # App sends every 5 seconds
    UNKNOWN_1 = bytes.fromhex('EFDD 0200 0000')
    HEARTBEAT = bytes.fromhex('EFDD 0002 0002 00')
    STATUS_REQUEST = bytes.fromhex('EFDD 0600 0000')  # seq can increment

    # Both UNKNOWN_1 and STATUS_REQUEST seem to increment a sequence numbber
    # after every command send
    # ... 0204 0400 and 0604 0400
    # ... 0205 0500 and 0605 0500
    # wraps at ff back to 00

    TARE        = bytes.fromhex('EFDD 0400 0000')

# Messages that don't conform to the MessageType|length format
# The 'efdd 200c 0c00' pattern seems to occur when reconnecting
# to the scale after it has been off a while.

NONSTANDARD_MESSAGES = (
    bytes.fromhex('efdd 2008 0800'),    # Power up?
    bytes.fromhex('efdd 200c 0c00'),    # Power down?
    bytes.fromhex('efdd 200a 0a00'),    # ?
)


def hex_logstr(message: Union[bytes, bytearray]) -> str:
    """
    Utility to render as groups of 4 nibbles, from the left
    """
    return message.hex(bytes_per_sep=-2, sep=' ')


class TimestampedData (NamedTuple):
    timestamp: float
    data: Union[bytes, bytearray]

# This holds the common functionality, but is NOT registered
class AcaiaGeneric (GenericScale):

    # _supports_prefixes = None
    _requires_heartbeat = True
    _heartbeat_period = 5  # Seconds
    _style = InteractionStyle.CLASSIC

    def __init__(self):
        super(AcaiaGeneric, self).__init__()
        self._adopt_sync()

    def _adopt_sync(self):
        # This config override is for development only, will be deprecated
        # 8<
        if config.acaia.INTERACTION_STYLE is not None:
            self._style = InteractionStyle(config.acaia.INTERACTION_STYLE)
        if config.acaia.REQUIRES_HEARTBEAT is not None:
            self._requires_heartbeat = config.acaia.REQUIRES_HEARTBEAT
        if config.acaia.HEARTBEAT_PERIOD is not None:
            self._heartbeat_period = config.acaia.HEARTBEAT_PERIOD
        # >8

        self._nominal_period = 0.1  # seconds per sample
        self._minimum_tare_request_interval = 2.5 * self._nominal_period
        self._sensor_lag = 0.38  # seconds, including all transit delays
        self._tare_timeout = 1.0  # seconds until considered coincidence
        self._tare_threshold = 0.05  # grams, within this, considered "at zero"

        self._info: Optional[Info] = None

        # logger = pyDE1.getLogger(self.__class__.__name__)
        # logger_notify = logger.getChild('notify')
        self._packet_buffer = bytearray()
        self._packet_buffer_lock = asyncio.Lock()

        self._heartbeat: Optional[SupervisedTask] = None
        self._control_lock = asyncio.Lock()

        self._setting_seq_number = 0
        self._timer_seq_number = 0

        if self._style == InteractionStyle.CLASSIC:
            self._command_cuuid = CUUID.AGE
            self._notify_cuuid = CUUID.AGE
            self._notify_lock = self._control_lock
        else:
            self._command_cuuid = CUUID.COMMAND
            self._notify_cuuid = CUUID.WEIGHT
            self._notify_lock = asyncio.Lock()

    async def _adopt_class(self):
        self._adopt_sync()

    async def _leave_class(self):
        try:
            self._heartbeat.cancel()
        except AttributeError as e:
            logger.error(f"Unable to cancel Acaia heartbeat: {e}")
        for attr in (
            # Class attributes, not instance attributes
            # '_style',
            # '_requires_heartbeat',
            # '_heartbeat_period',
            '_info',
            '_packet_buffer',
            '_packet_buffer_lock',
            '_heartbeat',
            '_control_lock',
            '_setting_seq_number',
            '_timer_seq_number',
            '_command_cuuid',
            '_notify_cuuid',
            '_notify_lock',
        ):
            delattr(self, attr)

    async def _initialize_after_connection(self, hold_ready=False):
        await super(AcaiaGeneric, self)._initialize_after_connection(
            hold_ready=True)
        # Check here to confirm "really connected"
        # Another symptom is NOT_REALLY_CONNECTED received as a packet
        if not self._bleak_client.services.get_characteristic(
                self._command_cuuid.value):
            logger.warning(
                f"CUUID {self._command_cuuid.value} not found, "
                "disconnecting for reconnect")
            logger.info(
                ', '.join(map(lambda c: c.uuid.removesuffix(
                    '-0000-1000-8000-00805f9b34fb'),
                    list(self._bleak_client.services.characteristics.values()))))
            await self.disconnect(for_reconnect=True)
            return

        await self._send_ident()
        await self._send_config()
        if self._requires_heartbeat:
            self._heartbeat = SupervisedTask(
                self._send_heartbeat,
            )
            self._heartbeat.work.name=f"{self.__class__.__name__}_Heartbeat"

        self._notify_ready()

    async def _send_packet(self, packet: Union[bytes, bytearray]):
        async with self._control_lock:
            await self._bleak_client.write_gatt_char(
                self._command_cuuid.value, packet)

    async def _send_ident(self):
        await self._send_packet(FixedMessage.IDENT.value)

    async def _send_config(self):
        await self._send_packet(
            pack_message(MessageType.EVENT,
                         bytes.fromhex('0001 0102 0205 0304')))

    async def _send_heartbeat(self):
        more = ''
        while self.is_connected:  # would error if not connected
            try:
                await self._send_packet(FixedMessage.HEARTBEAT.value)
            except bleak.exc.BleakError as e:
                more = e
                break
            await asyncio.sleep(self._heartbeat_period)
        logger.info(f"Discontinuing heartbeat {more}")

    async def start_sending_weight_updates(self):
        async with self._notify_lock:
            await self._bleak_client.start_notify(
                self._notify_cuuid.value,
                self._notification_handler)

    async def stop_sending_weight_updates(self):
        async with self._notify_lock:
            await self._bleak_client.stop_notify(self._notify_cuuid.value)

    async def disconnect(self, for_reconnect=False):
        if self._heartbeat:
            await self._heartbeat.work.cancel()
            self._heartbeat = None
        await super(AcaiaGeneric, self).disconnect(for_reconnect=for_reconnect)

    async def display_on(self):
        pass

    async def display_off(self):
        pass

    async def _tare_internal(self):
        await self._send_packet(FixedMessage.TARE.value)

    # ACKs on timer events appear to reflect the sequence number
    # efdd 0c04 0b01 2105 2c
    #             || sequence number
    #                || 21 - stop, 01 - start, 41 - reset
    # Sometimes longer
    # efdd 0c0b 0b00 0105 2d00 0000 0202 123b
    # efdd 0c0b 0b01 2105 2e00 0000 0202 135c
    # efdd 0c0b 0b02 0105 3400 0000 0202 1442
    # efdd 0c0b 0b03 4105 3500 0000 0202 1583
    # efdd 0c0b 0b04 0105 3700 0000 0202 1645
    # efdd 0c0b 0b05 2105 3600 0000 0202 1764
    # efdd 0c0b 0b06 4105 3800 0000 0202 1886
    # efdd 0c0b 0b07 4105 3500 0000 0202 1983
    # efdd 0c0b 0b08 0105 3100 0000 0202 1a3f
    # efdd 0c0b 0b09 4105 2200 0000 0202 1b70

    # TIMER event -  efdd 0c05 0700 0502 070c
    #                             0:05?

    async def timer_start(self):
        seq = self._timer_seq_number
        self._timer_seq_number = (self._timer_seq_number + 1) & 0xff
        await self._send_packet(
            pack_request(MessageType.TIMER, seq, TimerRequestType.START.value))

    async def timer_stop(self):
        seq = self._timer_seq_number
        self._timer_seq_number = (self._timer_seq_number + 1) & 0xff
        await self._send_packet(
            pack_request(MessageType.TIMER, seq, TimerRequestType.STOP.value))

    async def timer_reset(self):
        seq = self._timer_seq_number
        self._timer_seq_number = (self._timer_seq_number + 1) & 0xff
        await self._send_packet(
            pack_request(MessageType.TIMER, seq, TimerRequestType.RESET.value))

    async def set_grams(self):
        seq = self._setting_seq_number
        self._setting_seq_number = (self._setting_seq_number + 1) & 0xff
        await self._send_packet(pack_config(ConfigType.UNITS,
                                            seq,
                                            ConfigUnits.G))

    async def set_ounces(self):
        seq = self._setting_seq_number
        self._setting_seq_number = (self._setting_seq_number + 1) & 0xff
        await self._send_packet(pack_config(ConfigType.UNITS,
                                            seq,
                                            ConfigUnits.OZ))

    async def set_beep(self, audible: bool):
        seq = self._setting_seq_number
        self._setting_seq_number = (self._setting_seq_number + 1) & 0xff
        await self._send_packet(pack_config(ConfigType.BEEP,
                                            seq,
                                            ConfigBeep.ON if audible else
                                                ConfigBeep.OFF))

    async def _notification_handler(self, sender: int, data: bytearray):
        """Callback for arriving BLE packets"""
        tsd = TimestampedData(timestamp=time.time(), data=data)
        async with self._packet_buffer_lock:
            # TODO: Watchdog this
            await self._process_data_under_lock(tsd)

    async def _process_data_under_lock(self, tsd: TimestampedData):
        """
        With self._packet_buffer_lock in hand...

        Messages can be split across multiple packets, as well as
        multiple messages arriving in a single packet.

        This passes single, "complete" messages to process_message()

        Seen with Lunar and perhaps not with Lunar 2001
            Scale.Acaia: Waiting for more bytes, 3 < 5 bytes ef:dd:0c
        """
        self._packet_buffer.extend(tsd.data)

        while (lpb := len(self._packet_buffer)):

            if lpb < 5:
                break

            # Check for the troublesome 'efdd 2008 0800' 'efdd ...'
            # for nsm in NONSTANDARD_MESSAGES:
            #     if self._packet_buffer.startswith(nsm):
            #         self._process_message(nsm)
            #         self._packet_buffer \
            #             = self._packet_buffer[len(nsm):]
            #         break

            if self._packet_buffer.startswith(b'\xef\xdd\x20'):
                await self._process_message(self._packet_buffer[0:6])
                self._packet_buffer = self._packet_buffer[6:]
                continue

            idx = self._packet_buffer.find(HEADER)
            if idx == 0:  # HEADER at start of buffer
                try:
                    len_byte = self._packet_buffer[3]
                except IndexError:
                    logger.debug(
                        f"Waiting for more bytes, no length byte yet")
                    break
                if lpb >= (loa := len_byte + 5):
                    if lpb > 26:  # 26 seen with ACK and TIMER
                        logger.warning(
                            f"Packet buffer getting long, at bytes: {lpb}")
                    await self._process_message(self._packet_buffer[0:loa],
                                                timestamp=tsd.timestamp)
                    self._packet_buffer = self._packet_buffer[loa:]
                else:
                    logger.debug(
                        f"Waiting for {loa - lpb} more bytes")
                    break
            else:
                if idx == -1:
                    if lpb == 1 \
                            and self._packet_buffer[0] == HEADER[0]:
                        logger.info(
                            "Packet buffer is just first HEADER byte")
                        break
                    if lpb >= 2:
                        discarded = self._packet_buffer
                        self._packet_buffer = bytearray()
                else:
                    discarded = self._packet_buffer[0:idx]
                    self._packet_buffer \
                        = self._packet_buffer[idx:]
                logger.warning(
                    "Packet buffer does not start with header, discarding "
                    + hex_logstr(discarded))



# Sometimes the Lunar connects from a Bluetooth perspective
# but doesn't respond as one would expect. One symptom of this
# is when the needed CUUID is missing, another is when
# 'efdd 0707 0219 0100 0501 2108' is received as a packet.
#
# Rather than wait for the Lunar to drop the connection from its end,
# drop from this side and initiate a reconnection.

# TODO: Think through this more carefully
#           Should a disconnect be called if already reconnecting?
#           When should the _reconnect_task be cleared?

    async def _process_message(self, message: Union[bytearray, bytes],
                               timestamp: Optional[float] = None):
        if not timestamp:
            timestamp = time.time()

        if message == NOT_REALLY_CONNECTED:
            logger.error("NOT REALLY CONNECTED packet seen")
            # "NOT REALLY CONNECTED packet seen, disconnecting to reconnect")
            # Strange, sometimes this continues without issue. Watch it
            # await self.disconnect(for_reconnect=True)

        if message == b'\xef\xdd\x20\x08\x08\x00':
            logger_notify.info("Acaia 0x2008 = power up?")
            return
        elif message == b'\xef\xdd\x20\x0a\x0a\x00':
            logger_notify.info("Acaia 0x200a = power down?")
            return
        elif message == b'\xef\xdd\x20\x0c\x0c\x00':
            logger_notify.info("Acaia 0x200c = power down?")
            return

        try:
            length_byte = message[3]
            expected_length = 3 + length_byte + 2
            if len(message) != expected_length:
                logger_notify.error(
                    f"Expected {expected_length} packet, "
                    f"got {len(message)} bytes: {hex_logstr(message)}")
        except IndexError:
            logger_notify.error(
                f"Very short packet, "
                f"got {len(message)} bytes: {hex_logstr(message)}")
            return

        if len(message) < 7:
            logger_notify.error(f"Short packet: {hex_logstr(message)}")
            return

        try:
            message_type = MessageType(message[2])
        except ValueError as e:
            logger_notify.error(
                f"0x{message[2]:02x} {e}: {hex_logstr(message)}")
            return

        if message_type == MessageType.EVENT:
            try:
                event_type = EventType(message[4])
            except ValueError as e:
                logger_notify.error(f"{e}: {hex_logstr(message)}")
                return
        else:
            event_type = None  # Not really needed, but keeps pyCharm happy

        #

        if message_type == MessageType.EVENT:

            if event_type == EventType.WEIGHT:

                # Common for all three variants

                # 6 bytes or more bytes before checksum
                mantissa = message[5] + message[6] * 256 + message[7] * 65536
                scale_by = 10 ** message[9]  # Production do this faster
                if message[10] & 0x02:
                    sign = -1
                else:
                    sign = 1
                weight = sign * mantissa / scale_by

                if (message[10] & 0x01):  # Weight unsettled if & 0x01
                    other = '~'
                else:
                    other = ''

                if length_byte == 0x08:
                    pass

                elif length_byte == 0x0a:
                    # Seen on Lunar but not Lunar 2021
                    unknown11 = message[11]
                    battery = message[12]
                    unknown13 = message[13]
                    other = f"{other} {battery}% {unknown11}[11] {unknown13}[13]"
                    logger_notify.debug(
                        f"0x0a length: {weight}g {other} {hex_logstr(message)}")

                elif length_byte == 0x0c:
                    # it is a status, weight notification, "long version"

                    unk11 = message[11]
                    minutes = message[12]
                    seconds = message[13]
                    tenths = message[14]  # Seemingly, though why "2" at start?

                    other = f"{other} {unk11} {minutes}:{seconds:02.0f},{tenths:01.0f}"

                elif length_byte == 0x0e:
                    # it is a status, weight notification, "longer version"

                    unk11 = message[11]
                    battery = message[12]  # Guessing, 0x64 at 100%
                    unknown = message[13]
                    minutes = message[14]
                    seconds = message[15]
                    tenths = message[16]  # Seemingly, though why "2" at start?

                    other = f"{other} {unknown} " \
                            f"{minutes}:{seconds:02.0f},{tenths:01.0f} " \
                            f"- {unk11} {battery}%"

                else:
                    logger_notify.error(
                        f"{message_type.name}, {event_type.name} "
                        f"0x{len(message) - 4:02x} bytes unexpected: "
                        f"{hex_logstr(message)}")

                # asyncio.get_running_loop().create_task(
                await self.event_weight_update.publish(
                        ScaleWeightUpdate(
                            arrival_time=timestamp,
                            scale_time=self._scale_time_from_latest_arrival(
                                timestamp),
                            weight=weight
                        ))
                # )

            elif event_type == EventType.REPLY_06:
                logger_notify.info(
                    f"{message_type.name}, {event_type.name}: "
                    f"{hex_logstr(message)}")

            elif event_type == EventType.TIMER:
                logger_notify.info(
                    f"{message_type.name}, {event_type.name}: "
                    f"{hex_logstr(message)}")

            elif event_type == EventType.KEY:
                logger_notify.info(
                    f"{message_type.name}, {event_type.name}: "
                    f"{hex_logstr(message)}")

                # KEY: 0c: 0a 08 08 05 09 00 00 00 02 03 1d
                # KEY: 0c: 0a 08 08 05 14 00 00 00 02 01 28
                # KEY: 0c: 0a 08 08 05 15 00 00 00 02 03 29
                # KEY: 0c: 0a 08 08 05 16 00 00 00 02 03 2a
                # KEY: 0c: 0a 08 08 05 17 00 00 00 02 03 2b
                # KEY: 0c: 0a 08 08 05 18 00 00 00 02 03 2c
                # KEY: 0c: 0a 08 08 05 19 00 00 00 02 03 2d
                # KEY: 0c: 0a 08 09 05 14 00 00 00 02 03 29
                # KEY: 0c: 0a 08 09 05 17 00 00 00 02 03 2c
                # KEY: 0c: 0a 08 09 05 17 00 00 00 02 03 2c
                # KEY: 0c: 0a 08 09 05 18 00 00 00 02 03 2d
                # KEY: 0c: 0a 08 09 05 39 00 00 00 02 03 4e
                # KEY: 0c: 0a 08 0a 05 00 00 00 00 02 01 16
                # KEY: 0c: 0a 08 0a 05 12 00 00 00 02 01 28
                # KEY: 0c: 0e 08 08 05 17 00 00 00 02 03 07
                # KEY: 0c: 0e 08 08 05 18 00 00 00 02 03 07
                # KEY: 0c: 0e 08 09 05 17 00 00 00 02 03 07
                # KEY: 0c: 0e 08 09 05 17 00 00 00 02 03 07

            elif event_type == EventType.ACK:
                logger_notify.info(
                    f"{message_type.name}, {event_type.name}: "
                    f"{hex_logstr(message)}")

        elif message_type == MessageType.TARE:
            logger_notify.info(
                f"{message_type.name}: {hex_logstr(message)}")

        elif message_type == MessageType.INFO:
            # Not connected? WARNING Notify: INFO: 07: 07 02 19 01 00 01
            logger_notify.info(
                f"{message_type.name}: {hex_logstr(message)}")

        # Clues to status-message byte assignments from
        # https://github.com/oscar-b/scales/blob/master/src/acaia/scale.ts#L160
        elif message_type == MessageType.STATUS:
            payload = message[4:-2]
            battery = payload[0]
            try:
                units = ConfigUnits(payload[1])
            except ValueError as e:
                logger.error(f"In processing STATUS, {e}")
                units = '?'
            unk2 = payload[2]
            try:
                auto_off = ConfigAutoOff(payload[3])
            except ValueError as e:
                logger.error(f"In processing STATUS, {e}")
                auto_off = '?'
            unk4 = payload[4]
            try:
                beep = ConfigBeep(payload[5])
            except ValueError as e:
                logger.error(f"In processing STATUS, {e}")
                beep = '?'
            try:
                range = ConfigRange(payload[7])
            except ValueError as e:
                logger.error(f"In processing STATUS, {e}")
                range = '?'

            level = logging.INFO
            try:
                self._info = Info(
                    battery=battery,
                    units=units,
                    unk2=unk2,
                    auto_off=auto_off,
                    unk4=unk4,
                    beep=beep,
                    range=ConfigRange(range),
                )
            except ValueError as e:
                level = logging.INFO
                logger_notify.error(f"Error saving info data {e}")

            if battery > 100:
                level = logging.ERROR
            logger_notify.log(level,
                              "{}: {}% {} ({}) {} ({}) {} {}".format(
                                  message_type.name,
                                  battery,
                                  units.name,
                                  unk2,
                                  auto_off.name,
                                  unk4,
                                  beep.name,
                                  range.name,
                              ))

        elif message_type == MessageType.IDENTIFY:
            logger_notify.info(
                f"{message_type.name}: {hex_logstr(message)}")

        elif message_type == MessageType.TIMER:
            logger_notify.info(
                f"{message_type.name}: {hex_logstr(message)}")

        else:
            logger_notify.warning(
                f"Unrecognized message type: {hex_logstr(message)}")


# Register later to not conflict with ACAIAL1
@register_scale_class
class AcaiaAcaia (AcaiaGeneric):
    _supports_prefixes = ['ACAIA', 'ACAIAL1']
    _requires_heartbeat = False
    _style = InteractionStyle.CLASSIC


@register_scale_class
class AcaiaLunar(AcaiaGeneric):
    # Original Lunar seems to be CLASSIC
    _supports_prefixes = ['LUNAR']
    _requires_heartbeat = False
    _style = InteractionStyle.LUNAR

    def __init__(self):
        super(AcaiaLunar, self).__init__()
        # Seems like same speed as Skale II
        # when Lunar on "fast" filtering (default)
        self._sensor_lag = 0.38


@register_scale_class
class AcaiaProch (AcaiaGeneric):
    _supports_prefixes = ['PROCH']
    _requires_heartbeat = False
    _style = InteractionStyle.CLASSIC


@register_scale_class
class AcaiaPearlS (AcaiaGeneric):
    _supports_prefixes = ['PEARLS']
    _requires_heartbeat = False
    _style = InteractionStyle.LUNAR


@register_scale_class
class AcaiaPyxis (AcaiaGeneric):
    _supports_prefixes = ['PYXIS']
    _requires_heartbeat = False
    _style = InteractionStyle.LUNAR



"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

# These need to be here (or somewhere)
# to get the implementations added to the lists
import pyDE1.scale.acaia
import pyDE1.scale.atomax_skale_ii
import pyDE1.scale.felicita_arc



"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import asyncio
import enum
import sys
import time

import pyDE1
from pyDE1.scale.events import ScaleWeightUpdate, ScaleButtonPress
from pyDE1.scale.generic_scale import register_scale_class, GenericScale

logger = pyDE1.getLogger('Scale.AtomaxSkaleII')


@register_scale_class
class AtomaxSkaleII (GenericScale):

    _supports_prefixes = ['Skale']

    def __init__(self):
        super(AtomaxSkaleII, self).__init__()
        self._adopt_sync()

    def _adopt_sync(self):
        self._nominal_period = 0.1  # seconds per sample
        self._minimum_tare_request_interval = 2.5 * self._nominal_period
        self._sensor_lag = 0.38  # seconds, including all transit delays
        self._tare_timeout = 1.0  # seconds until considered coincidence
        self._tare_threshold = 0.05  # grams, within this, considered "at zero"

        # Enable tare on button 1, hold UUID if need to unsubscribe
        self._button_1_tare_subscriber_id = None
        self._task_button_press = asyncio.create_task(
            self._subscribe_button_press())

        # Linux, at least on an RPi 3B, needs response=True for write_gatt_char
        self._write_gatt_char_response = (sys.platform == 'linux')

    async def _adopt_class(self):
        self._adopt_sync()

    async def _leave_class(self):
        try:
            self._task_button_press.cancel()
        except AttributeError:
            pass
        await self._event_button_press.unsubscribe(
            self._button_1_tare_subscriber_id)
        for attr in (
            '_write_gatt_char_response',
            '_button_1_tare_subscriber_id',
            '_task_button_press',
        ):
            delattr(self, attr)

    async def _initialize_after_connection(self, hold_ready=False):
        await super(AtomaxSkaleII, self)._initialize_after_connection(
            hold_ready=True)
        await self.set_grams()
        if not hold_ready:
            self._notify_ready()

    async def start_sending_weight_updates(self):
        await self._bleak_client.start_notify(
            Characteristic.WEIGHT_NOTIFY_EF81.cuuid,
            self._weight_update_handler)
        logger.info("Sending weight updates")

    async def stop_sending_weight_updates(self):
        await self._bleak_client.stop_notify(
            Characteristic.WEIGHT_NOTIFY_EF81.cuuid)
        logger.info("Stopped weight updates")

    def is_sending_weight_updates(self):
        return NotImplementedError

    async def _tare_internal(self):
        await self.send_command(Command.TARE)
        logger.info("Internal tare sent")

    async def current_weight(self):
        # await self._bleak_client.read_gatt_char(
        #     Characteristic.UNKNOWN_EF83.uuid)
        # # That CUUID doesn't look like weight
        return None

    async def display_on(self):
        await self.send_command(Command.DISPLAY_WEIGHT)
        await self.send_command(Command.DISPLAY_ON)
        logger.info("Display on")

    async def display_off(self):
        await self.send_command(Command.DISPLAY_OFF)
        logger.info("Display off")

    async def set_grams(self):
        await self.send_command(Command.GRAMS)
        logger.info("Grams selected")

    @property
    def supports_button_press(self):
        return True

    async def start_sending_button_updates(self):
        await self._bleak_client.start_notify(
            Characteristic.BUTTON_NOTIFY_EF82.cuuid,
            self._button_press_handler)
        logger.info("Sending button updates")

    async def stop_sending_button_updates(self):
        await self._bleak_client.stop_notify(
            Characteristic.BUTTON_NOTIFY_EF82.cuuid)
        logger.info("Stopped button updates")

    async def send_command(self, command: "Command"):
        await self._bleak_client.write_gatt_char(
            command.cuuid,
            command.value,
            response=self._write_gatt_char_response)

    async def _update_self_from_device(self):
        self._model_number = await self._bleak_client.read_gatt_char(
            Characteristic.MODEL_NUMBER.cuuid)
        self._fw_revision = await self._bleak_client.read_gatt_char(
            Characteristic.FW_REVISION.cuuid)
        self._hw_revision = await self._bleak_client.read_gatt_char(
            Characteristic.HW_REVISION.cuuid)
        self._sw_revision = await self._bleak_client.read_gatt_char(
            Characteristic.SW_REVISION.cuuid)
        self._manufacturer_name = await self._bleak_client.read_gatt_char(
            Characteristic.MANUFACTURER_NAME.cuuid)

    async def _weight_update_handler(self, sender, data):

        try:
            now = time.time()

            w1 = int.from_bytes(data[1:4], byteorder='little',
                                signed=True) / 10.0
            w2 = int.from_bytes(data[5:8], byteorder='little',
                                signed=True) / 10.0
            # if w1 == w2:
            #     print(f"{dt:8.6f} {w1}")
            # else:
            #     print(f"{dt:8.6f} {w1}  {w2}")

            self._update_scale_time_estimator(now)

            await self.event_weight_update.publish(
                ScaleWeightUpdate(
                    arrival_time=now,
                    scale_time=self._scale_time_from_latest_arrival(now),
                    weight=w1
                ))
        except Exception as e:
            logger.exception(e)
            raise e

    async def _button_press_handler(self, sender, data):

        now = time.time()
        b = data[0]
        sbp = ScaleButtonPress(arrival_time=now, button=b)
        await self.event_button_press.publish(sbp)

    async def _button_press_subscriber(self, sbp: ScaleButtonPress) -> None:
        if sbp.button == 1:
            await self.tare()
            logger.info("Button 1 - Tare requested")

    async def _subscribe_button_press(self):
        self._button_1_tare_subscriber_id \
            = await self._event_button_press.subscribe(
            self._button_press_subscriber)
        self._task_button_press = None


class Characteristic(enum.Enum):
    CONFIGURATION_EF80 = 'ef80'  # W
    WEIGHT_NOTIFY_EF81 = 'ef81'  # N
    BUTTON_NOTIFY_EF82 = 'ef82'  # N
    UNKNOWN_EF83 = 'ef83'  # R

    BATTERY_LEVEL = '2a19'  # 0-63

    MODEL_NUMBER = '2a24'  # "SkaleII"

    FW_REVISION = '2a26'
    HW_REVISION = '2a27'
    SW_REVISION = '2a28'
    MANUFACTURER_NAME = '2a29'  # "ATOMAX INC."

    UNKNOWN = '0f050002-3225-44b1-b97d-d3274acb29de'  # R

    @property
    def cuuid(self):
        return f"0000{self.value}-0000-1000-8000-00805f9b34fb"


# These typically get written to CONFIGURATION_EF80
class Command(enum.Enum):
    DISPLAY_WEIGHT = b'\xec'
    DISPLAY_ON = b'\xed'
    DISPLAY_OFF = b'\xee'

    OUNCES = b'\x02'
    GRAMS = b'\x03'
    PERSIST_UNITS = b'\04'

    TARE = b'\x10'

    TIMER_ZERO = b'\xd0'
    TIMER_START = b'\xd1'
    TIMER_STOP = b'\xd2'

    # Place cal weight on Skale, send command
    CALIBRATE_500G = b'\xca\x05'
    CALIBRATE_1000G = b'\xca\x0a'

    # Undocumented at this time
    FILTER_WEAK = b'\x09\x00'
    FILTER_DEFAULT = b'\x09\x01'
    FILTER_STRONG = b'\x09\x02'
    FILTER_STRONGER = b'\x09\x03'

    @property
    def cuuid(self):
        return Characteristic.CONFIGURATION_EF80.cuuid



# DO NOT EDIT -- This file automatically generated -- DO NOT EDIT
git_data = {
    "hash": "174df9c7",
    "subject": "WIP: Doc updates",
    "ref_names": "HEAD -> alpha-2023-01",
    "author_date": "Thu Dec 15 12:09:22 2022 -0800",
    "commit_date": "Sun Jan 22 17:19:48 2023 -0800",
}
source_data = {
    "git": git_data,
}



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import copy
import enum
import inspect
import json

import math  # for nan to be uniquely math.nan
import time
from functools import reduce
from typing import Union, Dict, Set

import pyDE1.scanner
from pyDE1 import scanner
from pyDE1.config import config
from pyDE1.de1 import DE1
from pyDE1.de1.c_api import PackedAttr, MMR0x80LowAddr, pack_one_mmr0x80_write
from pyDE1.de1.notifications import MMR0x80Data
from pyDE1.dispatcher.mapping import MAPPING, TO, IsAt
from pyDE1.dispatcher.payloads import APIRequest
from pyDE1.dispatcher.resource import Resource
from pyDE1.event_manager.event_manager import SubscribedEvent
from pyDE1.exceptions import (
    DE1APITypeError, DE1APIValueError, DE1APIAttributeError, DE1APIKeyError,
    DE1NotConnectedError
)
from pyDE1.flow_sequencer import FlowSequencer
from pyDE1.scale.processor import ScaleProcessor
from pyDE1.scanner import scan_from_api
from pyDE1.utils import prep_for_json
from pyDE1.utils_public import rgetattr, rsetattr

logger = pyDE1.getLogger('Inbound.Implementation')

"""
How this works:

http/run.py creates an APIRequest and adds it to the queue

The _request_queue_processor from dispatcher picks it up and calls
GET:        resource_dict = await get_resource_to_dict(got.resource)
PATCH, PUT: await patch_resource_from_dict(got.resource, got.payload)

The resource_dict becomes a payload for the APIResponse object that is queued

To be able to return meaningful results from PUT/PATCH, the recipient 
needs to know which of possibly many elements the result comes from.
As most won't be returning anything, at least right now, append each,
within a nested dict, to a list and pass that back.

2022-12
=======

Any time a PATCH or PUT (state-modifying request) is processed, 
an MQTT packet is sent out with the content of the impacted area.

At this time those areas include:
    * de1/profile/id
    * de1/control
    * de1/setting
    * de1/calibration
    
The retain flag was considered for these, but has not been added
as the broker might be down or have been down, so a retained packet
may not be available or may not represent current state.
"""


def get_timeout(prop, value):

    name = prop.__name__
    timeout = config.http.ASYNC_TIMEOUT

    if name == 'connectivity_setter' and value:
        timeout = config.bluetooth.CONNECT_TIMEOUT + 0.100 \
                  + config.http.ASYNC_TIMEOUT + 0.100

    elif name in ('change_de1_to_id',
                  'change_scale_to_id',
                  'change_to_id',):
        if value is None:
            timeout = config.bluetooth.DISCONNECT_TIMEOUT \
                      + config.http.ASYNC_TIMEOUT + 0.100
        elif value == 'scan':
            timeout = config.bluetooth.SCAN_TIME \
                      + config.bluetooth.DISCONNECT_TIMEOUT \
                      + config.bluetooth.CONNECT_TIMEOUT \
                      + config.http.ASYNC_TIMEOUT + 0.100
        else:
            timeout = config.bluetooth.DISCONNECT_TIMEOUT \
                      + config.bluetooth.CONNECT_TIMEOUT \
                      + config.http.ASYNC_TIMEOUT + 0.100

    elif name in ('set_profile_by_id', 'upload_json_v2_profile'):
        timeout = config.http.PROFILE_TIMEOUT

    elif name == 'stop_at_time_set_async':
        timeout = config.http.ASYNC_TIMEOUT * 2

    elif name == 'upload_firmware_from_content':
        timeout = config.http.FIRMWARE_TIMEOUT

    if DE1().uploading_firmware:
        timeout += config.de1.CUUID_LOCK_WAIT_TIMEOUT

    return timeout


async def _prop_value_setter(prop, value):

    if inspect.isroutine(prop):
        if inspect.iscoroutinefunction(prop):
            retval = await asyncio.wait_for(prop(value),
                                            get_timeout(prop, value))
        else:
            retval = prop(value)
    else:
        raise DE1APITypeError (f"Setter {prop} is not callable")

    return retval


async def _prop_value_getter(prop):

    if inspect.isroutine(prop):
        if inspect.iscoroutinefunction(prop):
            retval = await asyncio.wait_for(prop(),
                                            get_timeout(prop, None))
        else:
            retval = prop()
    else:
        raise DE1APITypeError (f"Setter {prop} is not callable")

    return retval


# NB: This assumes that the MMR and CUUID are kept up to date
#     and that those that are read don't change on their own

async def _get_isat_value(isat: IsAt):

    # TODO: if IsAt.use_getter is implemented in the future, change here

    target = isat.target
    attr_path = isat.attr_path
    if attr_path is None:
        raise DE1APIAttributeError(f"Write-only attribute {isat.__repr__()}")

    de1 = DE1()
    flow_sequencer = FlowSequencer()
    scale_processor = ScaleProcessor()
    scale = scale_processor.scale
    thermometer = flow_sequencer._steam_temp_controller._thermometer

    retval = None

    # For any attribute or property with a getter, getattr() "just works"
    # If possibly a callable, need to
    #     await _prop_value_getter(rgetattr(target, attr_path))

    if target == TO.DE1:
        retval = rgetattr(de1, attr_path)

    elif target == TO.FlowSequencer:
        retval = rgetattr(flow_sequencer, attr_path)

    elif target == TO.Scale:
        retval = rgetattr(scale, attr_path)

    elif target == TO.ScaleProcessor:
        retval = rgetattr(scale_processor, attr_path)

    elif target == TO.Thermometer:
        retval = rgetattr(thermometer, attr_path)

    elif isinstance(target, MMR0x80LowAddr):
        # NB: This assumes that the MMR and CUUID are kept up to date
        #     and that those that are read don't change on their own
        # 2022-08: .read_always attribute added to handle self-changing MMRs

        if not de1.is_ready:
            raise DE1NotConnectedError(
                "DE1 is not connected at last-chance check")

        if attr_path != '':
            raise DE1APIAttributeError(
                "MMR reads do not support attr_path "
                f"{isat}")

        # For now, assume everything is kept current

        # TODO: Unify with replication in PATCH operation

        if target.value > de1.feature_flag.last_mmr0x80:
            retval = None
            logger.info(
                f"Skipping (not in FW) {target.name}, "
                f"0x{target.value:04x} > 0x{de1.feature_flag.last_mmr0x80:04x}")
        else:
            try:
                retval = de1._mmr_dict[target].data_decoded
            except KeyError:
                retval = None
            if retval is None or target.read_always:
                t0 = time.time()
                # TODO: Can this be simplified/clarified?
                ready = await de1.read_one_mmr0x80(target)
                await ready.wait()
                retval = de1._mmr_dict[target].data_decoded
                t1 = time.time()
                logger.debug(
                    f"Read of {target.__repr__()} took \t"
                    f"{(t1 - t0) * 1000:6.1f} ms"
                )

    elif inspect.isclass(target) and issubclass(target, PackedAttr):
        # NB: This assumes that the MMR and CUUID are kept up to date
        #     and that those that are read don't change on their own

        if not de1.is_ready:
            raise DE1NotConnectedError(
                "DE1 is not connected at last-chance check")

        try:
            obj = (de1._cuuid_dict[target.cuuid]).last_value
        except KeyError:
            logger.info(
                f"No last value for {target.cuuid}"
            )
            obj = None
        if obj is None:
            t0 = time.time()
            obj = await de1.read_cuuid(target.cuuid)
            t1 = time.time()
            logger.debug(
                f"Read of {target} took \t{(t1 - t0) * 1000:6.1f} ms"
            )
        retval = rgetattr(obj, attr_path)

    else:
        raise DE1APITypeError(
            f"Mapping target of {target} is not recognized"
        )

    return prep_for_json(retval)


async def _get_mapping_to_dict(partial_dict: dict) -> dict:
    """
    Takes a "branch" of a mapping and
      * Fills in any IsAt values
      * Recursively calls itself if a dict
      * Passes any other values unmodified
    """

    if isinstance(partial_dict, IsAt):
        partial_dict = { None: partial_dict }

    if not isinstance(partial_dict, dict):
        raise DE1APITypeError(f"Expected a dict, not {type(partial_dict)}")

    retval = {}

    for k, v in partial_dict.items():
        if isinstance(v, IsAt):
            try:
                this_val = await _get_isat_value(v)
            except AttributeError:
                if config.http.PRUNE_EMPTY_NODES:
                    continue # Don't write the key's entry
                else:
                    this_val = math.nan
        elif isinstance(v, dict):
            this_val = await _get_mapping_to_dict(v)
            # Suppress aggregates with nothing to aggregate
            if len(this_val) == 0 and config.http.PRUNE_EMPTY_NODES:
                continue
        else:
            this_val = v
        retval[k] = this_val

    return retval

# TODO: set and get versions (no recollection of what this means)

def _get_target_sets_inner(partial_dict: dict,
                           dict_of_sets: Dict[
                               str, Set[Union[PackedAttr, MMR0x80LowAddr]]],
                           include_can_read: bool,
                           include_can_write: bool):
    """
    Takes a "branch" of a mapping and returns a dict with two keys,
    'PacketAttr' and 'MMR0x80LowAddr', each with a set of targets.
    The dict_of_sets is modified in-place.
    """

    # See also validate.py
    # Valid: dict with dict
    #        IsAt with byte, bytearray (profile or firmware)
    if isinstance(partial_dict, IsAt):
        partial_dict = { None: partial_dict }

    if not isinstance(partial_dict, dict):
        raise DE1APITypeError(f"Expected a dict, not {type(partial_dict)}")

    for k, isat in partial_dict.items():

        if isinstance(isat, IsAt):
            target = isat.target
            if not ((inspect.isclass(target) and issubclass(target, PackedAttr))
                    or isinstance(target, MMR0x80LowAddr)):
                continue
            writable = target.can_write \
                       and not (isat.read_only
                                or (isat.attr_path is None
                                    and isat.setter_path is None))
            readable = target.can_read and isat.attr_path is not None
            if ((include_can_read and readable)
                    or (include_can_write and writable)):
                if isinstance(target, MMR0x80LowAddr):
                    dict_of_sets['MMR0x80LowAddr'].add(target)
                elif inspect.isclass(target) and issubclass(target, PackedAttr):
                    dict_of_sets['PackedAttr'].add(target)
            else:
                if not readable and not writable:
                    logger.error(
                        "IsAt should have either or both "
                        f"attr_path or setter_path not None: {isat.__repr__()}"
                    )

        elif isinstance(isat, dict):
            _get_target_sets_inner(isat, dict_of_sets,
                                   include_can_read, include_can_write)


def get_target_sets(mapping: dict,
                    include_can_read=False, include_can_write=False) \
        -> Dict[str, Set[Union[MMR0x80LowAddr, PackedAttr]]]:

    retval = {
        'MMR0x80LowAddr': set(),
        'PackedAttr': set()
    }

    _get_target_sets_inner(mapping, retval,
                           include_can_read, include_can_write)
    return retval


async def get_resource_to_dict(resource: Resource) -> dict:

    mapping = MAPPING[resource]
    return await _get_mapping_to_dict(mapping)


# PATCH and PUT are related, but have slightly different requirements
#
# For PATCH, it is sufficient that each entry from the request has
# a corresponding entry in the overall mapping dictionary
#
# For PUT it is required that each element in the mappig dictionary
# have a corresponding entry in the request dictionary (completeness)
#

# TODO: How to handle read-only attributes in the PUT case is TBD

async def patch_resource_from_dict(resource: Resource, values_dict: dict):

    mapping = MAPPING[resource]

    # Get the list of PackedAttrs that need to be patched (properties and
    # MMRs are handled one at a time, as they are atomic), lock from changes,
    # get copies of current PackedAttrs to patch

    # TODO: Managing locks
    # TODO: PackedAttr "getter" that will wait on a pending update, if one is
    #       in flight, as well as unify the retrieval if not present

    # target_sets = get_target_sets(values_dict, include_can_write=True)

    # TODO: Optimize this -- this needs a mapping, the way it is written now
    #       It really should only retrieve those that are being changed
    #       in the case of a PATCH

    target_sets = get_target_sets(mapping, include_can_write=True)

    # Lock here
    de1 = DE1()
    flow_sequencer = FlowSequencer()

    for pa in target_sets['PackedAttr']:
        pa: PackedAttr
        cuuid = pa.cuuid
        try:
            last_value = de1._cuuid_dict[cuuid]._last_value
        except KeyError:
            last_value = None
        if last_value is None:
            t0 = time.time()
            last_value = await de1.read_cuuid(cuuid)
            t1 = time.time()
            logger.debug(
                f"Read of {cuuid} took \t"
                f"{(t1 - t0) * 1000:6.1f} ms"
            )

        # Don't load them all, just make sure they are there
        # Only load if they are being changed
        # pending_packed_attrs[pa] = last_value

    pending_packed_attrs = {}

    # Valid: dict with dict
    #        IsAt with byte, bytearray (profile or firmware)
    if isinstance(mapping, dict) \
            and isinstance(values_dict, dict):
        pass
    elif isinstance(mapping, IsAt) \
        and isinstance(values_dict, (bytes, bytearray, str)):
        # TODO: Unify this with validate.py
        if issubclass((t := mapping.internal_type), enum.Enum):
            values_dict = t(values_dict)
        # coerce into "standard form"
        mapping = { None: mapping }
        values_dict = { None: values_dict }
    else:
        raise DE1APITypeError(
            "Implementation: Mapping and patch inconsistent, "
            "dict with dict, IsAt with raw/str value "
            f"not {type(mapping)} with {type(values_dict)}"
        )

    results_list = list()
    await _patch_dict_to_mapping_inner(values_dict,
                                       mapping,
                                       pending_packed_attrs,
                                       list(),
                                       results_list)

    # if there are pending_packed_attrs, send them

    # Potentially gather, but may not be faster
    for pa in pending_packed_attrs.values():
        await de1.write_packed_attr(pa)

    # release locks

    return results_list


async def _patch_dict_to_mapping_inner(partial_value_dict: dict,
                                       partial_mapping_dict: dict,
                                       pending_packed_attrs: Dict[
                                           type(PackedAttr), PackedAttr],
                                       running_path: list,
                                       results_list: list):

    """
    This assumes that everything has been determined as "valid"
    """

    # TODO: PATCH Method for HTTP specifies that PATCH is to be atomic
    #       This isn't completely possible as there is no
    #       roll-back capability implemented at this time.
    #       To at least be somewhat compliant, all keys
    #       should be processed as fully as possible
    #       prior to writing any to their targets.
    #
    #       https://datatracker.ietf.org/doc/html/rfc5789
    #
    #       NB: Although this is similar to JSON Merge Patch
    #       https://datatracker.ietf.org/doc/html/rfc7386
    #       it is subtly different in that a JSON 'null'
    #       is interpreted as the valid, Python 'None'
    #       rather than as a request to delete an element:
    #
    #           This design means that merge patch documents are suitable for
    #           describing modifications to JSON documents that primarily
    #           use objects for their structure and do not make use of
    #           explicit null values.
    #
    #           The merge patch format is not appropriate for all JSON syntaxes.

    de1 = DE1()
    flow_sequencer = FlowSequencer()
    scale_processor = ScaleProcessor()
    scale = scale_processor.scale
    thermometer = flow_sequencer._steam_temp_controller._thermometer

    for key, new_value in partial_value_dict.items():

        try:
            mapping_isat = partial_mapping_dict[key]
        except KeyError:
            raise DE1APIKeyError(
                f"Unable to find mapping for {key} on the specified path"
            )

        if isinstance(mapping_isat, IsAt):
            target = mapping_isat.target
            attr_path = mapping_isat.attr_path
            setter_path = mapping_isat.setter_path

            if mapping_isat.read_only or attr_path is None and setter_path is None:
                raise DE1APIValueError(
                    f"Mapping for '{key}': {mapping_isat} is not writable"
                )

            if isinstance(target, TO):
                if target == TO.DE1:
                    this_target = de1
                elif target == TO.FlowSequencer:
                    this_target = flow_sequencer
                elif target == TO.Scale:
                    this_target = scale
                elif target == TO.ScaleProcessor:
                    this_target = scale_processor
                elif target == TO.Thermometer:
                    this_target = thermometer
                elif target == TO.Scanner:
                    this_target = scanner
                else:
                    raise DE1APITypeError(
                        f"Unsupported target for '{key}': {mapping_isat}")

                if setter_path is not None:
                    # Allow for a non-property setter to return a value
                    setter = rgetattr(this_target, setter_path)
                    retval = await _prop_value_setter(setter, new_value)
                    if retval is not None:
                        # reduce(lambda a, b: {b: a}, [5,4,3,2,1], 'val')
                        #     {1: {2: {3: {4: {5: 'val'}}}}}
                        result_dict = reduce(lambda a, b: {b: a},
                                             running_path,
                                             retval)
                        results_list.append(result_dict)

                else:
                    rsetattr(this_target, attr_path, new_value)

            # TODO: Is there a better way to work with an unbound function?
            #       Maybe attach it to a module, rathern than a special case?
            elif target is None:
                if setter_path == 'scan_from_api':
                    retval = await scan_from_api(new_value)
                    if retval is not None:
                        # reduce(lambda a, b: {b: a}, [5,4,3,2,1], 'val')
                        #     {1: {2: {3: {4: {5: 'val'}}}}}
                        result_dict = reduce(lambda a, b: {b: a},
                                             running_path,
                                             retval)
                        results_list.append(result_dict)

            elif isinstance(target, MMR0x80LowAddr):

                # MMR writes need to be serial and are atomic in that
                # each writable MMR is a single value.
                # As a result, just write it here and now.

                if attr_path != '':
                    raise DE1APIValueError(
                        "MMR writes do not support attr_path "
                        f"{target} {mapping_isat}")

                if setter_path is not None:
                    raise DE1APIValueError(
                        "MMR writes do not support setter_path "
                        f"{target} {mapping_isat}")

                mmr_write = pack_one_mmr0x80_write(
                    addr_low= target,
                    value= new_value,
                )

                logger.debug(f"MMR to be written: {mmr_write.as_wire_bytes()}")

                await de1.write_packed_attr(mmr_write)

                # TODO: Should this wait on ready.wait() ??
                #       Or is there a way to collect them all for later
                #       as a speed optimization?
                ns: MMR0x80Data = de1._mmr_dict[target]
                await ns.ready_event.wait()

            elif inspect.isclass(target) and issubclass(target, PackedAttr):
                # NB: This assumes that the CUUIDs are kept up to date
                #     and that those that are read don't change on their own

                packed_attr = (de1._cuuid_dict[target.cuuid]).last_value
                # TODO: Can this be sped up reliably?
                if packed_attr is None:
                    packed_attr = await de1.read_cuuid(target.cuuid)
                old_value = rgetattr(packed_attr, attr_path)
                if new_value != old_value:
                    if not target in pending_packed_attrs:
                        pending_packed_attrs[target] = copy.deepcopy(packed_attr)
                    rsetattr(pending_packed_attrs[target], attr_path, new_value)

                # Send in outer method once all nodes are visited

            else:
                raise DE1APITypeError(
                    f"Mapping target of {target} is not recognized")

        elif isinstance(mapping_isat, dict):
            # TODO: Where did "this_val" (unused) come from?
            this_val = await _patch_dict_to_mapping_inner(
                partial_value_dict[key],
                partial_mapping_dict[key],
                pending_packed_attrs,
                # Prepend as reduce needs "reversed" list and need copy anyway
                running_path=[mapping_isat] + running_path,
                results_list=results_list)
        else:
            this_val = new_value


async def generate_mqtt_push(req: APIRequest):
    logger.debug(f"Generating readback for {req}")
    resource_path_changed = req.resource.value
    resource_area = None
    for res in (Resource.DE1_PROFILE_ID,
               Resource.DE1_SETTING,
               Resource.DE1_CONTROL,
               Resource.DE1_CALIBRATION,
               ):
        if resource_path_changed.startswith(res.value):
            resource_area = res
            break
    if resource_area is None:
        return
    timestamp = time.time()
    read_back_dict = await get_resource_to_dict(resource_area)
    if 'timestamp' not in read_back_dict.keys():
        read_back_dict['timestamp'] = timestamp
    read_back_dict['subtopic'] = resource_area.value
    SubscribedEvent.outbound_pipe.send(json.dumps(read_back_dict))


"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

# Only one scan can be running at the same time, at least with bleak 0.19
# on BlueZ. Using an exclusive lock can lead to pileups. Should be able
# to support a use case where the user finishes with a choose operation
# and then moves on to the next before the timeout expires.
#
# In a multi-client situation, one user canceling another user's activity
# isn't a great experience, but seems a reasonable compromise here.
#
# Intended behavior is that any new scan cancels any currently running one.
# The cancelled scan sends its scanning:false packet and the new scan sends
# an empty packet to "reset" clients.

import asyncio
import queue
import time

from typing import Iterable, Tuple, Optional

import bleak
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from pyDE1.event_manager.event_manager import SubscribedEvent
from pyDE1.event_manager.events import DeviceRole

import pyDE1
from pyDE1.config import config
from pyDE1.event_manager.payloads import EventPayload
from pyDE1.exceptions import *


logger = pyDE1.getLogger('Bluetooth.Scanner')

# Need an object with a class for the SubscribedEvent
class BluetoothScanner:
    pass

_scan_results_event = SubscribedEvent(BluetoothScanner())
scan_complete = asyncio.Event()


class _RegisteredPrefixes:

    def __init__(self):
        self._prefixes = dict()
        for role in DeviceRole:
            self._prefixes[role]: set[str] = set()
        self.add_to_role('', DeviceRole.UNKNOWN)

    def get_for_role(self, role: DeviceRole):
        if role is None:
            return frozenset()
        else:
            return frozenset(self._prefixes[role])

    def add_to_role(self, prefix: str, role: DeviceRole):
        self._prefixes[role].add(prefix)


RegisteredPrefixes = _RegisteredPrefixes()


# Can't just use nonlocal as it has to be bound in an outer function
# Getting "this module" is reasonably ugly
#   sys.modules[__name__]
#   __import__(__name__)

class _Wrapper:

    def __init__(self):
        self.scanner: Optional[bleak.BleakScanner] = None
        self.role: Optional[DeviceRole] = None
        self.found: list[BLEDevice] = list()
        self.timeout_task: Optional[asyncio.Task] = None

    async def new_run(self,
                scanner: bleak.BleakScanner,
                role: Optional[DeviceRole]=None,
                timeout: Optional[float]=None):
        if timeout is None:
            timeout = config.bluetooth.SCAN_TIME
        self.scanner = None
        self.role = None
        self.found = list()
        self.timeout_task = None
        scan_complete.clear()
        self.scanner = scanner
        self.role = role
        await _scan_results_event.publish(
                    ScanResults(_wrapper.found, role))
        await self.scanner.start()
        self.timeout_task = asyncio.create_task(self._stop_later(timeout))
        self.timeout_task.add_done_callback(
            lambda t: logger.info(
                f"Timeout task done: {t.get_name()}"
                + (" (cancelled)" if t.cancelled() else '')))
        logger.info(f"Scan beginning for {role} {self.timeout_task.get_name()}")

    async def end_run(self):
        # Cancel first, as was not seeming to cancel when done second.
        try:
            self.timeout_task.cancel()
            logger.info(f"Cancelled timeout: {self.timeout_task.get_name()}")
        except AttributeError:
            pass
        try:
            await self.scanner.stop()
            self.scanner = None
            await _scan_results_event.publish(
                ScanResults(self.found, self.role, scanning=False))
            scan_complete.set()
            logger.info(
                f"Scan ended for {self.role}, {len(self.found)} found")
        except AttributeError:
            pass

    async def _stop_later(self, after: float):
        my_scanner = self.scanner
        await asyncio.sleep(after)
        logger.info(
            f"Timeout, stopping scanner: {asyncio.current_task().get_name()}")
        if self.scanner == my_scanner:
            # This is a little dance as calling end_run() will cancel this task
            _wrapper.timeout_task = None
            await self.end_run()
        else:
            logger.warning(
                "Timeout, _stop_later: Stale task "
                + asyncio.current_task().get_name())


_wrapper = _Wrapper()


async def scan_until_timeout(role: DeviceRole = None,
                             timeout = None):

    if timeout is None:
        timeout = config.bluetooth.SCAN_TIME

    loop = asyncio.get_running_loop()

    prefix_set = RegisteredPrefixes.get_for_role(role)

    def check_match(device: BLEDevice, adv: AdvertisementData):
        for prefix in prefix_set:
            if adv.local_name and adv.local_name.startswith(prefix):
                logger.info(
                    f"'{prefix}' matched at {device.address} by {adv}")
                # Can get called more than once if the device data changes
                if device.address not in [d.address
                                          for d in _wrapper.found]:
                    _wrapper.found.append(device)
                loop.create_task(_scan_results_event.publish(
                    ScanResults(_wrapper.found, role)))

    # Cancel any already running
    await _wrapper.end_run()

    scanner = bleak.BleakScanner(detection_callback=check_match)

    try:
        await _wrapper.new_run(scanner, role, timeout)
    except bleak.exc.BleakDBusError as e:
        if e.args[0] == 'org.bluez.Error.InProgress':
            raise DE1OperationInProgressError(e)
        else:
            raise e


async def find_first_matching(role: DeviceRole,
                              timeout=None) -> BLEDevice:

    # Can't use .find_device_by_filter as it is a class method
    # and creates its own

    if timeout is None:
        timeout = config.bluetooth.SCAN_TIME

    loop = asyncio.get_running_loop()

    prefix_set = RegisteredPrefixes.get_for_role(role)

    found_event = asyncio.Event()

    def check_match(device: BLEDevice, adv: AdvertisementData):
        for prefix in prefix_set:
            if adv.local_name and adv.local_name.startswith(prefix):
                logger.info(
                    f"'{prefix}' matched at {device.address} by {adv}")
                _wrapper.found = [device]
                found_event.set()

    async def found_waiter():
        await found_event.wait()
        await _wrapper.end_run()

    # Cancel any already running
    await _wrapper.end_run()

    scanner = bleak.BleakScanner(detection_callback=check_match)

    task_found = asyncio.create_task(found_waiter())

    try:
        await _wrapper.new_run(scanner, role, timeout)
    except bleak.exc.BleakDBusError as e:
        if e.args[0] == 'org.bluez.Error.InProgress':
            raise DE1OperationInProgressError(e)
        else:
            raise e

    # Either found_water or timeout will set scan_complete
    await scan_complete.wait()
    task_found.cancel()
    try:
        return _wrapper.found[-1]
    except IndexError:
        return None



async def scan_from_api(role: DeviceRole):
    await scan_until_timeout(role)


class ScanResults (EventPayload):

    def __init__(self,
                 ble_device_list: list[BLEDevice],
                 role: DeviceRole,
                 scanning: bool = True):
        arrival_time = time.time()
        create_time = arrival_time
        super().__init__(arrival_time, create_time)
        self._version = "1.0.0"
        self.role: DeviceRole = role
        self.scanning = scanning
        if ble_device_list is not None:
            self.devices = [
                {'address': d.address, 'name': d.name, 'rssi': d.rssi}
                for d in ble_device_list ]
        else:
            self.devices = []



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Enum that defines the recognized resources that the Dispatcher can access
"""

import enum

RESOURCE_VERSION = '5.0.0'


class Resource (enum.Enum):

    SCAN = 'scan'
    SCAN_DEVICES = 'scan/devices'

    LOG = 'log/{id}'
    LOGS = 'logs'

    DE1 = 'de1'

    DE1_ID = 'de1/id'

    DE1_MODE = 'de1/mode'   # Commands
    DE1_STATE = 'de1/state' # Current state and substate

    DE1_FEATURE_FLAGS = 'de1/feature_flags'

    DE1_PROFILE = 'de1/profile'
    DE1_PROFILE_ID = 'de1/profile/id'
    DE1_PROFILE_STORE = 'de1/profile/store'  # PUT to database, but not DE1
    DE1_PROFILES = 'de1/profiles'

    DE1_FIRMWARE = 'de1/firmware'
    DE1_FIRMWARE_CANCEL = 'de1/firmware/cancel'
    DE1_FIRMWARE_ID = 'de1/firmware/id'
    DE1_FIRMWARES = 'de1/firmwares'

    DE1_AVAILABILITY = 'de1/availability'
    DE1_CONNECTIVITY = 'de1/connectivity'

    DE1_CONTROL = 'de1/control'
    DE1_CONTROL_ESPRESSO = 'de1/control/espresso'
    DE1_CONTROL_STEAM = 'de1/control/steam'
    DE1_CONTROL_HOT_WATER = 'de1/control/hot_water'
    DE1_CONTROL_HOT_WATER_RINSE = 'de1/control/hot_water_rinse'
    DE1_CONTROL_TANK_WATER_THRESHOLD = 'de1/control/tank_water_threshold'

    DE1_SETTING = 'de1/setting'
    DE1_SETTING_AUTO_OFF_TIME = 'de1/setting/auto_off_time'
    DE1_SETTING_FAN_THRESHOLD = 'de1/setting/fan_threshold'
    DE1_SETTING_START_FILL_LEVEL = 'de1/setting/start_fill_level'
    DE1_SETTING_BEFORE_FLOW = 'de1/setting/before_flow'
    DE1_SETTING_TARGET_GROUP_TEMP = 'de1/setting/target_group_temp'
    DE1_SETTING_USB_OUTLET = 'de1/setting/usb_outlet'
    DE1_SETTING_REFILL_KIT = 'de1/setting/refill_kit'

    DE1_SETTING_STEAM = 'de1/setting/steam'

    DE1_DEPRECATED = 'de1/setting/tbd_name_other_shot_settings'
    DE1_SETTING_TIME = 'de1/setting/time'

    DE1_CALIBRATION = 'de1/calibration'
    DE1_CALIBRATION_LINE_FREQUENCY = 'de1/calibration/line_frequency'
    DE1_CALIBRATION_FLOW_MULTIPLIER = 'de1/calibration/flow_multiplier'
    DE1_CALIBRATION_INTERNAL = 'de1/calibration/internal'

    DE1_PARAMETER_SET = 'de1/parameters'

    DE1_READ_ONCE = 'de1/read_once_values'

    DE1_PRESENCE = 'de1/presence'

    SCALE = 'scale'
    SCALE_ID = 'scale/id'

    SCALE_TARE = 'scale/tare'
    SCALE_DISPLAY = 'scale/display'

    SCALE_AVAILABILITY = 'scale/availability'
    SCALE_CONNECTIVITY = 'scale/connectivity'

    FLOW_SEQUENCER = 'flow_sequencer'
    FLOW_SEQUENCER_SETTING = 'flow_sequencer/settings'
    FLOW_SEQUENCER_PARAMETER_SET = 'flow_sequencer/parameter_set'

    THERMOMETER = 'thermometer'
    THERMOMETER_ID = 'thermometer/id'
    THERMOMETER_AVAILABILITY = 'thermometer/availability'


    VERSION = 'version'

    @property
    def can_get(self):
        retval = True

        # is FALSE
        if self in (
                self.SCAN,
                self.DE1_MODE,
                self.SCALE_TARE,
                self.SCALE_DISPLAY,
                # unimplemented
                self.DE1_PROFILE,
                self.DE1_PROFILE_STORE,
                self.DE1_PROFILES,
                self.DE1_FIRMWARE,
                self.DE1_FIRMWARES,
                self.DE1_DEPRECATED,
        ):
            retval = False

        return retval

    @property
    def can_put(self):
        retval = True

        # is FALSE
        if self in (
                self.SCAN_DEVICES,
                self.DE1_READ_ONCE,
                self.VERSION,
                self.LOG,
                self.LOGS,
                self.DE1_STATE,
                self.DE1_PROFILES,
                self.DE1_FIRMWARES,
                # unimplemented
                self.DE1_DEPRECATED,
        ):
            retval = False

        return retval

    @property
    def can_patch(self):
        retval = self.can_put
        # Can't PATCH firmware or profiles, have to PUT

        # is FALSE
        if self in (
                self.DE1_FIRMWARE,
                self.DE1_FIRMWARE_ID,
                self.DE1_PROFILE,
                self.DE1_PROFILE_ID,
                self.DE1_PROFILE_STORE,
                self.SCAN,
        ):
            retval = False

        return retval

    @property
    def can_post(self):
        retval = False
        # No POST implemented
        return retval

    @property
    def can_delete(self):
        retval = False
        # No DELETE implemented
        return retval


class ConnectivityEnum (enum.Enum):

    NOT_CONNECTED = 'not_connected'
    CONNECTED = 'connected'
    READY = 'ready'


class DE1ModeEnum (enum.Enum):

    SLEEP = 'Sleep'
    WAKE = 'Wake'
    STOP = 'Stop'
    END_STEAM = 'EndSteam'
    # TODO: Determine conditions under which these can be activated
    #       For now, assume the machine needs to be in Idle
    CLEAN = 'Clean'
    DESCALE = 'Descale'
    TRANSPORT = 'Transport' # AirPurge

    # Maybe sending NoRequest will trigger a report?
    NO_REQUEST = 'NoRequest'

    # Only valid during espresso flow
    SKIP_TO_NEXT = 'SkipToNext'

    # Only valid for non-GHC machines
    ESPRESSO = 'Espresso'
    STEAM = 'Steam'
    HOT_WATER = 'HotWater'
    HOT_WATER_RINSE = 'HotWaterRinse'




"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import time

from _operator import mul
from datetime import datetime
from typing import Optional, Union

import pyDE1
import pyDE1.shutdown_manager as sm

from pyDE1.config import config
from pyDE1.de1 import DE1
from pyDE1.de1.c_api import API_MachineStates
from pyDE1.de1.events import StateUpdate
from pyDE1.thermometer.bluedot import BlueDOT, BlueDOTUpdate
from pyDE1.utils import EventReadOnly


class SteamTempController:
    """
    Based on https://github.com/jeffsf/steam-to-temperature
    Rewritten for inclusion within pyDE1 directly

    This is not generalized to anything except the BlueDOT
    """

    def __init__(self, thermometer: BlueDOT):

        self.logger = pyDE1.getLogger('SteamTemp')

        self._thermometer = thermometer
        self._tat_estimator = TimeAtTargetEstimator()
        self._target: Optional[float] = None
        self._trigger_time = 0  # Cache between temperature updates

        self._de1 = DE1()

        self._on_trigger_event = asyncio.Event()
        self._on_trigger_event_ro = EventReadOnly(self._on_trigger_event)
        self._trigger_checker_task: Optional[asyncio.Task] = None

        loop = asyncio.get_running_loop()
        loop.create_task(
            self._thermometer.updates.subscribe(self._thermometer_subscriber))
        loop.create_task(
            self._de1.event_state_update.subscribe(self._de1_state_subscriber))

    @property
    def target(self):
        return self._tat_estimator.target

    @target.setter
    def target(self, value):
        self.logger.info(f"Target set: {value}")
        self._tat_estimator.target = value

    @property
    def has_triggered(self):
        return self._on_trigger_event.is_set()

    async def on_trigger_event(self):
        return self._on_trigger_event_ro

    # End of external API

    # The thermometer updates once per second, which isn't fast enough
    # to hit within one degree, especially for smaller volumes

    async def _trigger_checker(self):

        await asyncio.sleep(config.steam.SKIP_INITIAL_SECONDS)

        while (not self.has_triggered
               and self._de1.current_state == API_MachineStates.Steam
               and not sm.shutdown_underway.is_set()):
            now = time.time()
            if (self._trigger_time and
                    now >= self._trigger_time - config.steam.STOP_LAG):
                self._on_trigger_event.set()
                self.logger.info("Stopping steam with {} {} target".format(
                    self._target, self._thermometer.units.name))
                # TODO: Is this call up-to-date with firmware changes
                await self._de1.end_steam()
            else:
                # At 250 cal/sec, 100 g is 2.5°C / sec
                # 0.1 seconds should be well within 1°C
                await asyncio.sleep(0.1)
        if not self._on_trigger_event.set():
            self.logger.warning(
                "Trigger exited early for target {}, {},{}".format(
                    self._target,
                    self._de1.current_state, self._de1.current_substate
                ))
        self._trigger_checker_task = None
        await self.control_deactivate()

    async def control_activate(self):
        if not self._thermometer.is_ready:
            self.logger.error(
                f"{self._thermometer} is not ready, continuing anyway")
        if self._thermometer.is_ready:
            asyncio.create_task(self._thermometer.sample_fast())
        try:
            self._trigger_checker_task.cancel()
            self.logger.warning(
                "control_activate() called when active, reinitializing")
        except (AttributeError, asyncio.CancelledError):
            pass
        self._trigger_checker_task = None
        self._trigger_time = None
        self._on_trigger_event.clear()
        await self._tat_estimator.reset()
        self._trigger_checker_task = asyncio.create_task(
            self._trigger_checker(), name='TriggerChecker')
        self.logger.info("Control activated")

    async def control_deactivate(self):
        self.logger.info("No longer being controlled")
        self._trigger_time = None
        self._on_trigger_event.clear()
        await self._tat_estimator.reset()
        try:
            self._trigger_checker_task.cancel()
        except (AttributeError, asyncio.CancelledError):
            pass
        self._trigger_checker_task = None
        if self._thermometer.is_ready:
            await self._thermometer.sample_normal()

    async def _thermometer_subscriber(self, update: BlueDOTUpdate):

        # Use the BlueDOT high-alarm as the steam-to target

        self._target = update.high_alarm

        if self._trigger_checker_task is not None:
            self._trigger_time = await self._tat_estimator.new_sample(
                sample_time=update.arrival_time,
                temperature=update.temperature
            )
            if self._trigger_time:
                self.logger.info(
                    "TAT: {} in {:.3f} sec, {:.1f} °/s "
                    "at {:.1f} est {:.1f} rpt".format(
                        datetime.fromtimestamp(
                            self._trigger_time).time().isoformat(
                            timespec='milliseconds'),
                        self._trigger_time - time.time(),
                        self._tat_estimator.rate_of_rise,
                        self._tat_estimator.current_est,
                        update.temperature
                    )
                )

    async def _de1_state_subscriber(self, update: StateUpdate):
        if self._thermometer.address in ('', None):
            return

        if update.state == API_MachineStates.Sleep:
            if not self._thermometer.is_released:
                self.logger.info("DE1 reported Sleep, releasing thermometer")
                if self._thermometer.is_ready:
                    await self._thermometer.sample_slow()
                await self._thermometer.request_release()
        else:
            if not self._thermometer.is_ready:
                self.logger.info(
                    f"DE1 reported {update.state.name},{update.substate.name}, "
                    "capturing thermometer")
                await self._thermometer.request_capture()


class TimeAtTargetEstimator:
    """
    Class that encapsulates being able to estimate
    when the rate of rise will reach the target
    """

    logger = pyDE1.getLogger('SteamTemp.TAT')

    def __init__(self):
        self._history_time = []
        self._history_temperature = []
        self._history_lock = asyncio.Lock()
        self.big_gap = 2.1  # seconds, if gap exceeds, reset estimator
        self.target = None
        self._bhat = None
        self._mhat = None

    async def reset(self):
        async with self._history_lock:
            self._reset_have_lock()

    async def new_sample(self, sample_time: float,
                         temperature: Union[int, float]) -> Optional[float]:
        """
        Call when a new sample arrives

        Returns the time at target estimate as absolute time.time() reference
        """
        async with self._history_lock:
            if self._history_available:
                if (dt := sample_time - self._history_time[-1]) > self.big_gap:
                    self.logger.warning(
                        f"Resetting history after {dt:.1f} sec gap")
                    self._reset_have_lock()
            self._history_time.append(sample_time)
            while (len(self._history_time)
                    > config.steam.MAX_SAMPLES_FOR_ESTIMATE):
                self._history_time.pop(0)
            self._history_temperature.append(temperature)
            while (len(self._history_temperature)
                    > config.steam.MAX_SAMPLES_FOR_ESTIMATE):
                self._history_temperature.pop(0)

            return self._estimate_time_at_target_have_lock()

    async def estimate_time_at_target(self) -> Optional[float]:
        """
        Returns the time at target estimate as reference
        on same timescale as samples are reported
        """
        async with self._history_lock:
            return self._estimate_time_at_target_have_lock()

    @property
    def rate_of_rise(self):
        return self._mhat

    @property
    def current_est(self):
        return self._bhat

    def _reset_have_lock(self):
        self._history_time = []
        self._history_temperature = []
        self._mhat = None
        self._bhat = None
        self.logger.info("Reset history")

    @property
    def _history_available(self):
        # Ultra safe
        return min(len(self._history_temperature), len(self._history_time))

    def _estimate_time_at_target_have_lock(self) -> Optional[float]:
        # Use a least-squares estimator
        ns = self._history_available
        if ns < 2:
            t_target = None
        else:
            # Keep time scaled reasonably
            t0 = self._history_time[-1]
            t_norm = [t - t0 for t in self._history_time[-ns:]]
            s_x = sum(t_norm)
            s_y = sum(self._history_temperature[-ns:])
            s_xx = sum([x * x for x in t_norm])
            # s_yy = sum([y * y for y in self._history_temperature[-ns:]])
            s_xy = sum(map(mul,
                           t_norm,
                           self._history_temperature[-ns:]))
            self._mhat = (ns * s_xy - (s_x * s_y)) \
                   / (ns * s_xx - (s_x * s_x))
            if self._mhat:
                self._bhat = (s_y / ns) - self._mhat * (s_x / ns)
                t_target = t0 + (self.target - self._bhat) / self._mhat
            else:
                self._bhat = None
                t_target = None

        return t_target



"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Class to provide logging of waiting for locks

Typical usage:

        ll = LockLogger(some_lock, 'Name of lock').check()
        async with some_lock:
            ll.acquired()
            await something_under_some_lock()
        ll.released()
"""

import asyncio
import time

import pyDE1
from pyDE1.utils import call_str


class LockLogger:

    def __init__(self, lock: asyncio.Lock, name: str):
        self._lock = lock
        self._name = name
        self._logger = pyDE1.getLogger(f'Lock.{self._name}')
        self._checked = None
        self._acquired = None

    def check(self) -> 'LockLogger':
        if self._lock.locked():
            self._checked = time.time()
            self._logger.warning(
                f"Waiting for lock {repr(self._lock)} {call_str()}")
        else:
            self._checked = None
        return self  # Allows lock_logger = LockLogger(...).check()

    def acquired(self, full_trace=False):
        self._acquired = time.time()
        if self._checked:
            dt = (self._acquired - self._checked) * 1000
            # Warning here allows setting log level to warning
            # and still seeing how long the waits were
            self._logger.warning(
                f"Acquired lock after {dt:.0f} ms {call_str(full_trace)}")
        else:
            self._logger.debug(
                f"Acquired lock {call_str(full_trace)}")

    def released(self, full_trace=False):
        dt = (time.time() - self._acquired) * 1000
        if self._lock.locked():
            self._logger.error(
                f"NOT RELEASED after {dt:.0f} ms {call_str(full_trace)}")
        else:
            self._logger.info(
                f"Released lock after {dt:.0f} ms {call_str(full_trace)}")





"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import pprint
import time
import warnings
from typing import Optional, Union

import aiosqlite
from bleak.backends.device import BLEDevice

import pyDE1
from pyDE1.bledev.managed_bleak_device import (
    ManagedBleakDevice, ClassChanger, class_changer_generic_class,
)
from pyDE1.config import config
from pyDE1.dispatcher.resource import ConnectivityEnum
from pyDE1.event_manager.event_manager import SubscribedEvent
from pyDE1.event_manager.events import ConnectivityChange, DeviceRole
from pyDE1.exceptions import (
    DE1NoAddressError,
    DE1RuntimeError, DE1NotConnectedError,
    DE1UnsupportedDeviceError,
)
from pyDE1.scale.events import ScaleWeightUpdate, ScaleTareSeen
from pyDE1.scanner import RegisteredPrefixes

logger = pyDE1.getLogger('Scale')

# Used for class selection and for BLE detection and filtering
_prefix_to_class = dict()


def register_scale_class(cls: 'GenericScale'):
    # logger.warning("This warning doesn't appear")
    if isinstance(cls._supports_prefixes, str):
        # This may not appear in the logs as during import sequence
        logger.error(
            "{}._supports_prefixes should be a list-like iterable, "
            "not a string '{}'".format(cls, cls._supports_prefixes))
        warnings.warn(
            "{}._supports_prefixes should be a list-like iterable, "
            "not a string '{}'".format(cls, cls._supports_prefixes),
            category=RuntimeWarning,
            stacklevel=2
        )
        cls._supports_prefixes = list(cls._supports_prefixes)
    for prefix in cls._supports_prefixes:
        _prefix_to_class[prefix] = cls
        if prefix not in (None, ''):
            RegisteredPrefixes.add_to_role(prefix, DeviceRole.SCALE)
    return cls


def prefix_to_class(prefix: Optional[str]):
    if prefix is None:
        prefix = ''
    cls = None
    try:
        cls = _prefix_to_class[prefix]
    except KeyError:
        if prefix == '':
            raise DE1RuntimeError(
                "Missing empty-string key in _prefix_to_class: "
                f"{_prefix_to_class}")
        for key in _prefix_to_class.keys():
            if key != '' and prefix.startswith(key):
                cls = _prefix_to_class[key]
    if cls is None:
        raise DE1UnsupportedDeviceError(
            f"No recognized scale registered for '{prefix}'")
    return cls


# TODO: Experimentaly confirm that weight and mass-flow estimates
#       are reasonably time aligned - NB: DE1.fall_time

# TODO: Think about how to manage a "tare seen" event
#       and what the use cases for it would be.
#       Could use asyncio.Event(), but what are the states
#       and how do you "release" if it never arrives?

@register_scale_class
@class_changer_generic_class
class GenericScale (ClassChanger, ManagedBleakDevice):
    """
    Rework of Scales -- Generic should be the class initialized,
    changing address with a BLEDevice or connecting may then change
    the class, but shouldn't alter references to the GenericScale API

    NB: Device-specific references may come and go over address changes
    """

    _supports_prefixes = ('',)

    def __init__(self):

        self._role = DeviceRole.SCALE
        self.logger = pyDE1.getLogger('Scale.Generic')
        super(GenericScale, self).__init__()

        self._event_weight_update: SubscribedEvent = SubscribedEvent(self)
        self._event_button_press: SubscribedEvent = SubscribedEvent(self)
        self._event_tare_seen: SubscribedEvent = SubscribedEvent(self)
        self._event_scale_changed: SubscribedEvent = SubscribedEvent(self)

        self._adopt_sync()
        config.logging.handlers.STDERR = 'DEBUG'
        config.logging.formatters.STDERR = config.logging.formatters.LOGFILE
        self._period_estimator = PeriodEstimator(self)

        # Don't need to await this on instantiation
        asyncio.get_event_loop().create_task(
            self._event_weight_update.subscribe(self._self_callback))

    def _adopt_sync(self):
        """
        The "settings" for GenericScale -- split out for reuse in __init__
        """
        self._name: Optional[str] = None

        self._adjust_name_send_scale_change()
        self.logger = pyDE1.getLogger('Scale.Generic')

        # These are often model-specific, override in subclass init
        self._nominal_period = 0.1  # seconds per sample
        self._minimum_tare_request_interval = 2.5 * self._nominal_period
        self._sensor_lag = 0.38  # seconds, including all delays to arrival
        # From https://www.youtube.com/watch?v=SIzFhnZ32Y0
        # (James Hoffmann) at 4:51
        #   Hiroia    0.20
        #   Skale     0.33
        #   Felicita  0.45
        #   Acaia     0.64
        self._tare_timeout = 1.0  # seconds until considered coincidence
        self._tare_threshold = 0.05  # grams, within this, considered "at zero"
        self.hold_at_tare = False

        self._estimated_period = self._nominal_period
        self._last_weight_update_received = 0
        self._last_tare_request_sent = 0

        self._tare_requested = False

        try:
            self._period_estimator.reset(self._nominal_period)
        except AttributeError:
            pass

    async def _adopt_class(self):
        self._adopt_sync()

    @property
    def event_scale_changed(self):
        return self._event_scale_changed

    @property
    def sensor_lag(self):
        return self._sensor_lag

    def _adjust_name_send_scale_change(self):
        # logger.debug(f"Adjusting name {logger.findCaller(stacklevel=2)}")
        try:
            ble_name = self._bleak_client._backend._device_info['Name']
        except (KeyError, AttributeError, TypeError) as e:
            ble_name = '(unknown)'
        self._name = f"{self.__class__.__name__}: {ble_name}"
        sc = ScaleChange(arrival_time=time.time(),
                         state=self.connectivity_state,
                         id=self.address,
                         name=self.name)
        asyncio.create_task(self._event_scale_changed.publish(sc))

    async def _initialize_after_connection(self, hold_ready=False):
        # Check that this is the right class to service the connected device
        self._adjust_name_send_scale_change()
        ble_name = self._bleak_client._backend._device_info['Name']
        cls = prefix_to_class(ble_name)
        if type(self) != cls:
            self.logger.info(
                "Changing class on {} from {} to {} after connection".format(
                    self._name, type(self), cls))
            await self._change_class(cls)
            self._adjust_name_send_scale_change()
        await self.display_on()
        await self.start_sending_weight_updates()
        if self.supports_button_press:
            await self.start_sending_button_updates()
        await self._restore_period_from_db()
        if not hold_ready:
            self._notify_ready()

    async def connect(self):
        """
        TODO: Can this do a "first if found"
            if self.address in (None, '') ??
        """
        # if self.address in (None, ''):
        #     device = await find_first_matching(
        #         recognized_scale_prefixes())
        #     if device:
        #         await self.scale.change_address(device)
        #     else:
        #         raise DE1NoAddressError(
        #             f"Can't connect without an address")
        await self.capture()

    async def disconnect(self):
        await self.release()

    async def start_sending_weight_updates(self):
        raise DE1NotConnectedError

    async def stop_sending_weight_updates(self):
        raise DE1NotConnectedError

    @property
    def is_sending_weight_updates(self):
        raise DE1NotConnectedError

    @property
    def supports_button_press(self):
        return False

    async def start_sending_button_updates(self):
        raise DE1NotConnectedError

    async def stop_sending_button_updates(self):
        raise DE1NotConnectedError

    async def tare(self):
        """
        A tare request can only be made every
        self._minimum_tare_request_interval seconds

        It doesn't make sense to hammer it as it will take
        at least one reporting period to "see" the tare
        """
        dt = time.time() - self._last_tare_request_sent
        if dt > self._minimum_tare_request_interval:
            await self._tare_internal()
            self._last_tare_request_sent = time.time()
            self._tare_requested = True
            logger.info(f"Tare request sent")
        else:
            logger.info(
                f"Tare request skipped, too soon, {dt:0.3f} seconds")
        return self._last_tare_request_sent

    async def _tare_internal(self):
        raise DE1NotConnectedError

    async def current_weight(self) -> Optional[float]:
        """
        Intended to request an in-the-moment read from the scale
        If not supported, may return None instead of a weight
        """
        raise DE1NotConnectedError

    async def display_on(self):
        raise DE1NotConnectedError

    async def display_off(self):
        raise DE1NotConnectedError

    # The two *_bool for API

    async def tare_with_bool(self, do_it=True):
        if do_it:
            await self.tare()

    async def display_bool(self, on: bool):
        if on:
            await self.display_on()
        else:
            await self.display_off()

    @property
    def estimated_period(self):
        return self._estimated_period

    @property
    def event_weight_update(self):
        return self._event_weight_update

    @property
    def event_button_press(self):
        return self._event_button_press

    @property
    def event_tare_seen(self):
        return self._event_tare_seen

    async def change_address(self, address: Optional[Union[BLEDevice, str]]):
        """
        Change address, including changing type if passed a BLEDevice

        Raises DE1UnsupportedDeviceError if BLEDevice.name is not recognized
        """
        if isinstance(address, BLEDevice) or address in (None, ''):
            if address in (None, ''):
                cls = prefix_to_class('')
            else:
                # This call potentially raises DE1UnsupportedDeviceError
                cls = prefix_to_class(address.name)
            if type(self) != cls:
                self.logger.info(
                    "Address change for {} requested, "
                    "changing class from {} to {}".format(
                        address, type(self), cls))
                if not self.is_released:
                    # TODO: How should a timeout here be handled?
                    #       Should the timeout be configurable?
                    await self.release(
                        timeout=self._bleak_client._backend._timeout)
                await self._change_class(cls)
        changed = await self._bleak_client.change_address(address)
        if changed:
            self._adjust_name_send_scale_change()

    def _scale_time_from_latest_arrival(self,
                                        latest_arrival: float):
        """
        Given the latest arrival, provide "best" estimate
        of when that weight was on the scale

        At present, just compensates for scale._scale_delay
        which should include transit delays and the like
        """
        return latest_arrival - self._sensor_lag

    def _update_scale_time_estimator(self,
                                     latest_arrival:float):
        """
        Call once per arrival to update any "fancy" algorithms such as PLL
        """
        pass

    async def _self_callback(self, swu: ScaleWeightUpdate) -> None:
        dt = swu.arrival_time - self._last_weight_update_received
        self._last_weight_update_received = swu.arrival_time

        # TODO: Run profiler and evaluate if creating a task
        #       is consuming too much time

        asyncio.create_task(
            self._period_estimator.process_arrival(dt))

        if self._tare_requested:
            dt = swu.arrival_time - self._last_tare_request_sent
            if dt > self._tare_timeout:
                self._tare_requested = False
                logger.error(f"No tare seen after {dt:0.03f} seconds")
            elif abs(swu.weight) < self._tare_threshold:
                self._tare_requested = False
                await self.event_tare_seen.publish(
                    ScaleTareSeen(swu.arrival_time)
                )
                logger.info(f"Tare seen after {dt:0.03f} seconds")

        if self.hold_at_tare:
            if abs(swu.weight) > self._tare_threshold:
                # Timing will be checked in scale.tare()
                await self.tare()

    @property
    def nominal_period(self):
        return self._nominal_period

    @nominal_period.setter
    def nominal_period(self, value):
        self._nominal_period = value
        self._period_estimator.reset(value)

    async def _persist_period_to_db(self):
        if not self.address:
            raise DE1NoAddressError(
                "Can't persist scale period without a scale address")
        async with aiosqlite.connect(config.database.FILENAME) as db:
            sql = "INSERT OR REPLACE INTO persist_hkv " \
                  "(header, key, value) " \
                  "VALUES " \
                  "(:header, :key, :value) "
            await db.execute(sql, {
                'header': 'scale.period',
                'key': self.address,
                'value': self.estimated_period,
            })
            await db.commit()

    async def _restore_period_from_db(self):
        if not self.address:
            raise DE1NoAddressError(
                "Can't restore scale period without a scale address")
        async with aiosqlite.connect(config.database.FILENAME) as db:
            sql = "SELECT value FROM persist_hkv " \
                  "WHERE header = :header AND key = :key"
            cur = await db.execute(sql, {
                'header': 'scale.period',
                'key': self.address,
            })
            row = await cur.fetchone()
            if row and row[0]:
                val = float(row[0])
                logger.info(
                    "Loading scale-period estimate of "
                    f"{val:.5f} from database")
                self._estimated_period = val
                self._period_estimator.reset(val)
            else:
                logger.info(
                    "No previous scale-period estimate for "
                    f"{self.address} found")

    # For API
    @property
    def connectivity(self):
        retval = ConnectivityEnum.NOT_CONNECTED
        if self.is_connected:
            if self._ready.is_set():
                retval = ConnectivityEnum.READY
            else:
                retval = ConnectivityEnum.CONNECTED
        return retval


class ScaleChange(ConnectivityChange):
    """
    Gets sent when the address of the scale changes "behind the scenes"
    such as with a call to scale.take_over_from()

    Not sent on initialization at this time
    """
    pass


# Used by Scale instances
class PeriodEstimator:
    """
    Estimate inter-arrival period from stream of arrivals

    Presently just an exponential moving average

    Skale II usually "bulks up" two or more reports on a 150-ms clock
    300 ms burbles aren't uncommon. A "normal" 50-ms stretch before its
    other half arrives would generate a (50/100) * k change.
    The other half then would generate (-100/100) * k change
    So k on the order of 1/1000 should be reasonable (10 sec, ~1 min settle)
    k of 1/10000 would be even better (100 sec, 10 min settle)
    Another way to look at this is 600 ms error / 600 s measurement ~ 0.1%

    Hand-in-hand with this is how long to consider a gap vs. a burble
    Nearly 5% of 150-ms windows from a SkaleII had 3 reports.
    Up to 6 in a window were observed. It dropped to 0.1% at 4 reports
    per window. Ignoring too many of these can lead to the estimate being off.
    Based on this, 300 ms (two periods) seems too short.
    300 + 150/2 = 375 ms is probably OK.
    450 + 150/2 = 525 ms is probablu conservative
    Try 500 ms to be reasonable.
    """

    def __init__(self, my_scale):

        # TODO: How to update this PeriodEstimator for subclass changes?

        self._scale = my_scale

        self._k = 1/10000   # tau ~ 17 min at 10 samples/sec
        self._ma = self._scale.nominal_period
        self._too_long = 0.5  # seconds before considered a gap

        self._persist_every_n = 1000  # about 100 seconds
        self._n_counter = 0

    def reset(self, nominal_period: float):
        self._ma = nominal_period
        self._scale._estimated_period = nominal_period

    async def process_arrival(self, delta_arrival_time: float):

        if delta_arrival_time < self._too_long:
            self._ma = ((1 - self._k) * self._ma) \
                       + (self._k * delta_arrival_time)
            self._scale._estimated_period = self._ma
            self._n_counter += 1
            if self._n_counter >= self._persist_every_n:
                self._n_counter = 0
                logger.getChild('Period').debug(f"Persisting {self._ma}")
                await self._scale._persist_period_to_db()





"""
Copyright © 2021, 2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import asyncio
import logging
import multiprocessing
import os
from socket import gethostname

import paho.mqtt.client as mqtt
from paho.mqtt.client import MQTTv5, MQTT_CLEAN_START_FIRST_ONLY

import pyDE1
import pyDE1.shutdown_manager as sm
from pyDE1.config import config
from pyDE1.api.outbound.mqtt.run import MQTTStatusText

def attach(subtopic: str,
           loop: asyncio.AbstractEventLoop,
           logger: logging.Logger = None):

    try:
        pname = multiprocessing.current_process().name
    except AttributeError:
        pname = __name__
    if logger is None:
        logger = pyDE1.getLogger(pname)

    client_logger = logger.getChild('MQTTClient')
    client_logger.setLevel(logging.INFO)

    will_topic = config.mqtt.TOPIC_ROOT + '/' + subtopic

    class ClientSendsGracefulDisconnect (mqtt.Client):

        def disconnect(self, *args, **kwargs):
            _send_on_graceful_disconnect()
            super(ClientSendsGracefulDisconnect,
                  self).disconnect(*args, **kwargs)

    mqtt_client = ClientSendsGracefulDisconnect(
        client_id="{}@{}[{}]{}".format(
            config.mqtt.CLIENT_ID_PREFIX,
            gethostname(),
            os.getpid(),
            pname,
        ),
        clean_session=None,  # Required for MQTT5
        userdata=None,
        protocol=MQTTv5,
        transport=config.mqtt.TRANSPORT,
    )

    def _send_on_connection():
        client_logger.info(
            f"Publishing {will_topic} {MQTTStatusText.on_connection.value}")
        mqtt_client.publish(
            topic=will_topic,
            payload=MQTTStatusText.on_connection.value,
            qos=0,
            retain=True,
        )

    def _send_on_graceful_disconnect():
        client_logger.info(
            f"Publishing {will_topic}  {MQTTStatusText.on_graceful_disconnect.value}")
        rc = mqtt_client.publish(
            topic=will_topic,
            payload=MQTTStatusText.on_graceful_disconnect.value,
            qos=0,
            retain=True,
        )
        rc.wait_for_publish()

    def on_connect_callback(client, userdata, flags, reasonCode, properties):
        client_logger.log(logging.INFO if reasonCode == 0
                          else logging.ERROR,
                          f"CB: Connect: flags: {flags}, reasonCode: {reasonCode}, "
                          f"properties {properties}")
        if reasonCode == 0:
            _send_on_connection()
        else:
            client_logger.critical(
                f"Connection to MQTT broker failed: {str(reasonCode)}, ")
            # loop.call_soon_threadsafe(_start_shutdown)

    def on_disconnect_callback(client, userdata, reasonCode, properties=None):
        if sm.shutdown_underway.is_set():
            level = logging.INFO
        else:
            level = logging.ERROR
        client_logger.log(level, f"CB: Disconnect: reasonCode: {reasonCode}, "
                                 f"properties {properties}")

    mqtt_client.on_connect = on_connect_callback
    mqtt_client.on_disconnect = on_disconnect_callback
    mqtt_client.enable_logger(client_logger)

    if config.mqtt.USERNAME is not None:
        logger.info(f"Connecting MQTT with username '{config.mqtt.USERNAME}'")
        mqtt_client.username_pw_set(
            username=config.mqtt.USERNAME,
            password=config.mqtt.PASSWORD
        )

    mqtt_client.will_set(topic=will_topic,
                         payload=MQTTStatusText.on_will.value,
                         qos=0,
                         retain=True,
                         properties=None
                         )

    if config.mqtt.TLS:
        mqtt_client.tls_set(ca_certs=config.mqtt.TLS_CA_CERTS,
                            certfile=config.mqtt.TLS_CERTFILE,
                            keyfile=config.mqtt.TLS_KEYFILE,
                            cert_reqs=config.mqtt.TLS_CERT_REQS,
                            tls_version=config.mqtt.TLS_VERSION,
                            ciphers=config.mqtt.TLS_CIPHERS)

    async def cleanup_on_shutdown():
        # "Independent" shutdown for potential extraction
        client_logger.info("Watching for shutdown event")
        await sm.wait_for_shutdown_underway()
        client_logger.info("Shutting down MQTT client")
        if mqtt_client.is_connected():
            mqtt_client.disconnect()
        mqtt_client.loop_stop()
        client_logger.info("MQTT loop stopped")
        # This is a secondary process, don't set "done"

    loop.create_task(cleanup_on_shutdown())

    mqtt_client.connect(host=config.mqtt.BROKER_HOSTNAME,
                        port=config.mqtt.BROKER_PORT,
                        keepalive=config.mqtt.KEEPALIVE,
                        bind_address="",
                        bind_port=0,
                        clean_start=MQTT_CLEAN_START_FIRST_ONLY,
                        properties=None)

    mqtt_client.loop_start()


